#!/usr/bin/env python
import rich.console
import pandas as pd
import os
import re
import openpyxl
import sys
import json
import difflib
import inspect

import relecov_tools.utils
import relecov_tools.assets.schema_utils.jsonschema_draft
import relecov_tools.assets.schema_utils.metadatalab_template
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule
from datetime import datetime
from openpyxl.worksheet.datavalidation import DataValidation

pd.set_option("future.no_silent_downcasting", True)

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class SchemaBuilder(BaseModule):
    def __init__(
        self,
        input_file=None,
        schema_base=None,
        draft_version=None,
        diff=False,
        output_dir=None,
        version=None,
        project=None,
        non_interactive=False,
    ):
        """
        Initialize the SchemaBuilder class. This class generates a JSON Schema file based on the provided draft version.
        It reads the database definition from an Excel file and allows customization of the schema generation process.
        """
        super().__init__(output_dir=output_dir, called_module=__name__)
        self.excel_file_path = input_file
        self.non_interactive = non_interactive
        # Validate input data
        if not self.excel_file_path or not os.path.isfile(self.excel_file_path):
            self.log.error("A valid Excel file path must be provided.")
            raise ValueError("A valid Excel file path must be provided.")
        if not self.excel_file_path.endswith(".xlsx"):
            self.log.error("The Excel file must have a .xlsx extension.")
            raise ValueError("The Excel file must have a .xlsx extension.")
        # No metadata is being processed so batch_id will be execution date
        self.set_batch_id(self.basemod_date)

        # Validate output folder creation
        if not output_dir:
            self.output_dir = relecov_tools.utils.prompt_create_outdir(
                path=None, out_dir=None
            )
        else:
            self.output_dir = os.path.abspath(output_dir)
            if not os.path.exists(self.output_dir):
                self.output_dir = relecov_tools.utils.prompt_create_outdir(
                    path=None, out_dir=output_dir
                )

        # Get version option
        if not version:
            # If not defined, then ask via prompt
            self.version = relecov_tools.utils.prompt_text(
                "Write the desired version using semantic versioning:"
            )
        self.version = version
        if not relecov_tools.utils.validate_semantic_version(self.version):
            raise ValueError("[red]Error: Invalid version format")

        # Get version option
        # Parse build-schema configuration
        self.build_schema_json_file = os.path.join(
            os.path.dirname(__file__), "conf", "build_schema_config.json"
        )

        if project is None:
            project = relecov_tools.utils.prompt_text("Write the desired project:")
        self.project = project

        available_projects = self.get_available_projects(self.build_schema_json_file)

        build_schema_config = ConfigJson(self.build_schema_json_file)
        config_data = build_schema_config.get_configuration(
            "projects"
        )  # Obtener solo la sección "projects"
        self.configurables = config_data.get("configurables", {})

        if self.project in available_projects:
            self.project_config = config_data.get(self.project, {})
        else:
            self.log.error(
                f"No configuration available for '{self.project}'. Available projects: {', '.join(available_projects)}"
            )
            stderr.print(
                f"[red]No configuration available for '{self.project}'. Available projects: {', '.join(available_projects)}"
            )
            sys.exit(1)

        # Validate show diff option
        if diff is False:
            self.show_diff = None
        else:
            self.show_diff = True

        # Validate json schema draft version
        if not draft_version and self.non_interactive:
            self.draft_version = "2020-12"
        else:
            self.draft_version = (
                relecov_tools.assets.schema_utils.jsonschema_draft.check_valid_version(
                    draft_version
                )
            )

        # Validate base schema
        if schema_base is not None:
            if relecov_tools.utils.file_exists(schema_base):
                self.base_schema_path = schema_base
            else:
                self.log.error(
                    f"[Error]Defined base schema file not found: {schema_base}."
                )
                stderr.print(
                    f"[Error]Defined base schema file not found: {schema_base}. Exiting..."
                )
                sys.exit(1)
        else:
            try:
                config_json = ConfigJson()
                relecov_schema = config_json.get_topic_data(
                    "json_schemas", "relecov_schema"
                )
                try:
                    self.base_schema_path = os.path.join(
                        os.path.dirname(os.path.realpath(__file__)),
                        "schema",
                        relecov_schema,
                    )
                    if not relecov_tools.utils.file_exists(self.base_schema_path):
                        self.log.error(
                            "[Error]Fatal error. Relecov schema were not found in current relecov-tools installation. Make sure relecov-tools command is functioning."
                        )
                        stderr.print(
                            "[Error]Fatal error. Relecov schema were not found in current relecov-tools installation. Make sure relecov-tools command is functioning. Exiting..."
                        )
                        sys.exit(1)
                    self.log.info(
                        "RELECOV schema successfully found in the configuration."
                    )
                    stderr.print(
                        "[green]RELECOV schema successfully found in the configuration."
                    )
                except FileNotFoundError as fnf_error:
                    self.log.error(f"Configuration file not found: {fnf_error}")
                    stderr.print(f"[red]Configuration file not found: {fnf_error}")
                    sys.exit(1)
            except KeyError as key_error:
                self.log.error(f"Configuration key error: {key_error}")
                stderr.print(f"[orange]Configuration key error: {key_error}")
                sys.exit(1)

    def validate_database_definition(self, json_data):
        """Validate the mandatory features and ensure:
        - No duplicate enum values in the JSON schema.
        - Date formats follow 'YYYY-MM-DD'.

        Args:
            json_data (dict): The JSON data representing the database definition.

        Returns:
            dict: A dictionary containing errors found, categorized by:
                - Missing features
                - Duplicate enums
                - Incorrect date formats
        """
        log_errors = {
            "missing_features": {},
            "duplicate_enums": {},
            "invalid_example_types": {},
            "invalid_date_formats": {},
        }

        mandatory_features = [
            "enum",
            "examples",
            "ontology_id",
            "type",
            "description",
            "classification",
            "label_name",
            "fill_mode",
            "required (Y/N)",
            "complex_field (Y/N)",
        ]

        # Iterate over properties in json_data
        for prop_name, prop_features in json_data.items():
            missing_features = [
                feature
                for feature in mandatory_features
                if feature not in prop_features
            ]
            if missing_features:
                log_errors["missing_features"][prop_name] = missing_features

            # Check for duplicate enum values
            if prop_features.get("enum"):
                if not pd.isna(prop_features["enum"]):
                    enum_values = prop_features["enum"].split("; ")
                    # Verify that enum has no duplicates
                    if len(enum_values) != len(set(enum_values)):
                        duplicates = [
                            value
                            for value in set(enum_values)
                            if enum_values.count(value) > 1
                        ]
                        log_errors["duplicate_enums"][prop_name] = duplicates

            # Check date format for properties with type=string and format=date
            if (
                prop_features["type"] == "string"
                and prop_features.get("format") == "date"
            ):
                example = prop_features.get("examples")
                if example:
                    if isinstance(example, datetime):
                        example = example.strftime("%Y-%m-%d")
                    if isinstance(example, str):
                        try:
                            datetime.strptime(example, "%Y-%m-%d")
                        except ValueError:
                            if prop_name not in log_errors["invalid_date_formats"]:
                                log_errors["invalid_date_formats"][prop_name] = []
                            log_errors["invalid_date_formats"][prop_name].append(
                                f"Invalid date format '{example}', expected 'YYYY-MM-DD'"
                            )

        # return log errors if any
        if any(log_errors.values()):
            stderr.print("[red]\t- Database Validation Failed")
            # Convert log_errors dictionary to DataFrame
            df_errors = pd.DataFrame(
                [
                    {
                        "Error Category": category,
                        "Field": field,
                        "Details": (
                            ", ".join(details) if isinstance(details, list) else details
                        ),
                    }
                    for category, errors in log_errors.items()
                    for field, details in errors.items()
                ]
            )

            # Save errors to file
            error_file_path = f"{self.output_dir}/schema_validation_errors.csv"
            df_errors.to_csv(error_file_path, index=False, encoding="utf-8")

            # Provide errors to user in rich table format:
            relecov_tools.utils.display_dataframe_to_user(
                name="Schema Validation Errors", dataframe=df_errors
            )
            stderr.print(f"\t- Log errors saved to:\n\t{error_file_path}")

            # Ask user whether to continue or stop execution
            if self.non_interactive or relecov_tools.utils.prompt_yn_question(
                "Errors found in database values. Do you want to continue? (Y/N)"
            ):
                pass
            else:
                return log_errors
        else:
            stderr.print("[green]\t- Database validation passed")

        # If no errors found
        return None

    def get_available_projects(self, json):
        """Get list of available software in configuration

        Args:
            json (str): Path to bioinfo configuration json file.

        Returns:
            available_software: List containing available software defined in json.
        """
        config = relecov_tools.utils.read_json_file(json)
        # available_software = list(config.keys())
        available_projects = list(config.get("projects", {}).keys())
        return available_projects

    def read_database_definition(self, sheet_id="main"):
        """Reads the database definition from an Excel sheet and converts it into JSON format.

        Args:
            sheet_id (str): The sheet name or ID in the Excel file to read from. Defaults to "main".

        Returns:
            json_data (dict): The JSON data representing the database definition.
        """
        caller_method = inspect.stack()[1][3]
        # Read excel file
        df = pd.read_excel(
            self.excel_file_path,
            sheet_name=sheet_id,
            na_values=["nan", "N/A", "NA", ""],
        )
        # Convert database to json format
        json_data = {}
        for row in df.itertuples(index=False):
            property_name = row[0]
            values = row[1:]
            json_data[property_name] = dict(zip(df.columns[1:], values))

        # Check json is not empty
        if len(json_data) == 0:
            self.log.error(f"{caller_method}{sheet_id}) No data found in xlsx database")
            stderr.print(
                f"{caller_method}{sheet_id}) [red]No data found in xlsx database"
            )
            sys.exit(1)

        # Perform validation of database content
        validation_out = self.validate_database_definition(json_data)
        if validation_out:
            sys.exit()
        else:
            return json_data

    def create_schema_draft_template(self):
        """
        Create a JSON Schema draft template based on the draft version.
        Available drafts: [2020-12]

        Returns:
            draft_template(dict): The JSON Schema draft template.
        """
        draft_template = (
            relecov_tools.assets.schema_utils.jsonschema_draft.create_draft(
                draft_version=self.draft_version, required_items=True
            )
        )
        return draft_template

    def standard_jsonschema_object(
        seschemalf, data_dict, target_key, remove_ontology=False
    ):
        """
        Create a standard JSON Schema object for a given key in the data dictionary.

        Args:
            data_dict (dict): The data dictionary containing the properties.
            target_key (str): The key for which to create the JSON Schema object.

        Returns:
            json_dict (dict): The JSON Schema object for the target key.
        """
        # For enum and examples, wrap the value in a list
        json_dict = {}

        # Function to handle NaN values
        def handle_nan(value):
            if pd.isna(value) or value in ["nan", "NaN", "None", "none"]:
                return ""
            return str(value)

        if target_key in ["enum", "examples"]:
            value = handle_nan(data_dict.get(target_key, ""))
            # if no value, json key won't be necessary, then avoid adding it
            if len(value) > 0:
                items = value.split("; ")
                if remove_ontology and target_key == "enum":
                    items = [re.sub(r"\s*\[.*?\]", "", item).strip() for item in items]
                json_dict[target_key] = items if target_key == "enum" else [value]
        elif target_key == "description":
            json_dict[target_key] = handle_nan(data_dict.get(target_key, ""))
        else:
            json_dict[target_key] = handle_nan(data_dict.get(target_key, ""))
        return json_dict

    # TODO: needs validation
    def complex_jsonschema_object(self, property_id, features_dict):
        """
        Create a complex (nested) JSON Schema object for a given property ID.

        Args:
            property_id (str): The ID of the property for which to create the JSON Schema object.
            features_dict (dict): A dictionary mapping database features to JSON Schema features.

        Returns:
            json_dict (dict): The complex JSON Schema object.
        """
        json_dict = {"type": "object", "properties": {}}

        # Read tab-dedicated sheet in excell database
        try:
            complex_json_data = self.read_database_definition(sheet_id=property_id)
        except ValueError as e:
            self.log.error(f"{e}")
            stderr.print(f"[yellow]{e}")
            return None

        # Add sub property items
        for sub_property_id, _ in complex_json_data.items():
            json_dict["properties"][sub_property_id] = {}
            complex_json_feature = {}
            for db_feature_key, json_key in features_dict.items():
                if json_key == "required":
                    continue
                feature_schema = self.standard_jsonschema_object(
                    complex_json_data[sub_property_id], db_feature_key
                )
                if feature_schema:
                    complex_json_feature[json_key] = feature_schema[db_feature_key]
            json_dict["properties"][sub_property_id] = complex_json_feature

        return json_dict

    def build_new_schema(self, json_data, schema_draft):
        """
        Build a new JSON Schema based on the provided JSON data and draft template.

        Parameters:
        json_data (dict): Dictionary containing the properties and values of the database definition.
        schema_draft (dict): The JSON Schema draft template.

        Returns:
            schema_draft (dict): The newly created JSON Schema.
        """
        # Fill schema header
        # FIXME: it gets 'relecov-tools' instead of RELECOV
        new_schema = schema_draft
        project_name = relecov_tools.utils.get_package_name()
        new_schema["$id"] = relecov_tools.utils.get_schema_url()
        new_schema["title"] = f"{project_name} Schema."
        new_schema["description"] = (
            f"Json Schema that specifies the structure, content, and validation rules for {project_name}"
        )
        new_schema["version"] = self.version

        # Fill schema properties
        try:
            # List of properties to check in the features dictionary (it maps values between database features and json schema features):
            #       key[db_feature_key]: value[schema_feature_key]
            mapping_features = {
                "enum": "enum",
                "examples": "examples",
                "ontology_id": "ontology",
                "type": "type",
                "options": "options",
                "description": "description",
                "classification": "classification",
                "label_name": "label",
                "fill_mode": "fill_mode",
                "required (Y/N)": "required",
                "submitting_lab_form": "header",
            }
            required_property_unique = []

            # Read property_ids in the database.
            #   Perform checks and create (for each property) feature object like:
            #       {'example':'A', 'ontology': 'B'...}.
            #   Finally this objet will be written to the new schema.
            for property_id, db_features_dic in json_data.items():
                schema_property = {}
                required_property = {}

                # Parse property_ids that needs to be incorporated as complex fields in json_schema
                if json_data[property_id].get("complex_field (Y/N)") == "Y":
                    complex_json_feature = self.complex_jsonschema_object(
                        property_id, mapping_features
                    )
                    if complex_json_feature:
                        schema_property["type"] = "array"
                        schema_property["items"] = complex_json_feature
                        schema_property["additionalProperties"] = False
                        schema_property["required"] = [
                            key for key in complex_json_feature["properties"].keys()
                        ]
                # For those that follows standard format, add them to json schema as well.
                else:
                    for db_feature_key, schema_feature_key in mapping_features.items():
                        # Verifiy that db_feature_key is present in the database (processed excel (aka 'json_data'))
                        if db_feature_key not in db_features_dic:
                            self.log.info(
                                f"Feature {db_feature_key} is not present in database ({self.excel_file_path})"
                            )
                            stderr.print(
                                f"[INFO] Feature {db_feature_key} is not present in database ({self.excel_file_path})"
                            )
                            continue
                        if (
                            "required" in db_feature_key
                            or "required" == schema_feature_key
                        ):
                            is_required = str(db_features_dic[db_feature_key])
                            if is_required != "nan":
                                required_property[property_id] = is_required
                        elif db_feature_key == "options":
                            options_value = str(
                                db_features_dic.get("options", "")
                            ).strip()
                            if options_value:
                                options_dict = {}
                                options_list = options_value.split(",")

                                for option in options_list:
                                    key_value = option.split(":")
                                    if len(key_value) == 2:
                                        key = key_value[0].strip()
                                        value = key_value[1].strip()
                                        try:
                                            if "." in value:
                                                value = float(value)
                                            else:
                                                value = int(value)
                                        except ValueError:
                                            pass
                                        options_dict[key] = value
                                schema_property.update(options_dict)
                        else:
                            std_json_feature = self.standard_jsonschema_object(
                                db_features_dic, db_feature_key, remove_ontology=False
                            )
                            if std_json_feature:
                                schema_property[schema_feature_key] = std_json_feature[
                                    db_feature_key
                                ]
                            else:
                                continue
                # Finally, send schema_property object to the new json schema draft.
                new_schema["properties"][property_id] = schema_property

                # Add to schema draft the recorded porperty_ids.
                for key, values in required_property.items():
                    if values == "Y":
                        required_property_unique.append(key)
            # TODO: So far it appears at the end of the new json schema. Ideally it should be placed before the properties statement.
            new_schema["required"] = required_property_unique
            grouped_anyof = {}

            for prop_id, prop_data in json_data.items():
                group = str(prop_data.get("conditional_required_group", "")).strip()
                if group and group.lower() != "nan":
                    grouped_anyof.setdefault(group, []).append(prop_id)
            anyof_rules = []
            for props in grouped_anyof.values():
                if len(props) >= 1:
                    anyof_rules.extend([{"required": [prop]} for prop in props])
                    for prop in props:
                        if prop in required_property_unique:
                            required_property_unique.remove(prop)
            if anyof_rules:
                new_schema["anyOf"] = anyof_rules

            # Return new schema
            return new_schema

        except Exception as e:
            self.log.error(f"Error building schema: {str(e)}")
            stderr.print(f"[red]Error building schema: {str(e)}")
            raise

    def verify_schema(self, schema):
        """
        Verify that the given schema adheres to the JSON Schema specification for the specified draft version.

        Args:
            schema (dict): The JSON Schema to be verified.

        Raises:
            ValueError: If the schema does not conform to the JSON Schema specification.
        """
        relecov_tools.assets.schema_utils.jsonschema_draft.check_schema_draft(
            schema, self.draft_version
        )

    def get_schema_diff(self, base_schema, new_schema):
        """
        Print the differences between the base schema and the newly generated schema.

        Args:
            base_schema (dict): The base JSON Schema to compare against.
            new_schema (dict): The newly generated JSON Schema to compare.

        Returns:
            bool: True if differences are found, False otherwise.
        """
        # Set diff input
        base_schema_lines = json.dumps(base_schema, indent=4).splitlines()
        new_schema_lines = json.dumps(new_schema, indent=4).splitlines()

        # Get diff lines
        diff_lines = list(
            difflib.unified_diff(
                base_schema_lines,
                new_schema_lines,
                fromfile="base_schema.json",
                tofile="new_schema.json",
            )
        )

        if not diff_lines:
            self.log.info(
                "No differences were found between already installed and new generated schema. Exiting. No changes made"
            )
            stderr.print(
                "[yellow]No differences were found between already installed and new generated schema. Exiting. No changes made"
            )
            return None
        else:
            self.log.info(
                "Differences found between the existing schema and the newly generated schema."
            )
            stderr.print(
                "[yellow]Differences found between the existing schema and the newly generated schema."
            )
            if self.show_diff:
                return self.print_save_schema_diff(diff_lines)
            else:
                return None

    def print_save_schema_diff(self, diff_lines=None):
        # Set user's choices
        choices = ["Print to standard output (stdout)", "Save to file", "Both"]
        diff_output_choice = (
            "Save to file"
            if self.non_interactive
            else relecov_tools.utils.prompt_selection(
                "How would you like to print the diff between schemes?:", choices
            )
        )
        if diff_output_choice in ["Print to standard output (stdout)", "Both"]:
            for line in diff_lines:
                print(line)
            return True
        if diff_output_choice in ["Save to file", "Both"]:
            diff_filepath = os.path.join(
                os.path.realpath(self.output_dir) + "/build_schema_diff.txt"
            )
            with open(diff_filepath, "w") as diff_file:
                diff_file.write("\n".join(diff_lines))
            self.log.info(f"[green]Schema diff file saved to {diff_filepath}")
            stderr.print(f"[green]Schema diff file saved to {diff_filepath}")
            return True

    # FIXME: Add version tag to file name
    def save_new_schema(self, json_data):
        """
        Save the generated JSON Schema to the output folder.

        Args:
            json_data (dict): The JSON Schema to be saved.

        Returns:
            bool: True if the schema was successfully saved, False otherwise.
        """
        try:
            path_to_save = f"{self.output_dir}/relecov_schema.json"
            with open(path_to_save, "w") as schema_file:
                json.dump(json_data, schema_file, ensure_ascii=False, indent=4)
            self.log.info(f"New JSON schema saved to: {path_to_save}")
            stderr.print(f"[green]New JSON schema saved to: {path_to_save} ")
            return True
        except PermissionError as perm_error:
            self.log.error(f"Permission error: {perm_error}")
            stderr.print(f"[red]Permission error: {perm_error}")
        except IOError as io_error:
            self.log.error(f"I/O error: {io_error}")
            stderr.print(f"[red]I/O error: {io_error}")
        except Exception as e:
            self.log.error(f"An unexpected error occurred: {str(e)}")
            stderr.print(f"[red]An unexpected error occurred: {str(e)}")
        return False

    # FIXME: overview-tab - FIX first column values
    # FIXME: overview-tab - Still need to add the column that maps to tab metadatalab
    def create_metadatalab_excel(self, json_schema):
        """
        Generates an Excel template file for Metadata LAB with four sheets:
        Overview, Metadata LAB, Data Validation, and Version History.

        Args:
            json_schema (dict): The JSON schema used to generate the template.
                                It should include properties and required fields.

        Returns:
            None: If an error occurs during the process.
        """
        try:
            # Retrieve existing files in the output directory
            output_files = os.listdir(self.output_dir)
            notes_control_input = (
                "Auto-generated update"
                if self.non_interactive
                else input(
                    "\033[93mEnter a note about changes made to the schema: \033[0m"
                )
            )
            # Identify existing template files
            template_files = [
                f for f in output_files if f.startswith("Relecov_metadata_template")
            ]
            if template_files:
                # Extract the latest version number from existing files
                latest_file = max(
                    template_files,
                    key=lambda x: (
                        re.search(r"v(\d+\.\d+\.\d+)", x).group(1)
                        if re.search(r"v(\d+\.\d+\.\d+)", x)
                        else "0"
                    ),
                )
                version_history = pd.DataFrame(
                    columns=["FILE_VERSION", "CODE", "NOTES CONTROL", "DATE"]
                )
                match = re.search(r"v(\d+\.\d+\.\d+)", latest_file)
                if match:
                    # Load the latest template file and attempt to read version history
                    out_file = os.path.join(self.output_dir, latest_file)
                    try:
                        wb = openpyxl.load_workbook(out_file)
                        if "VERSION" in wb.sheetnames:
                            ws_version = wb["VERSION"]
                            data = ws_version.values
                            columns = next(data)
                            version_history = pd.DataFrame(data, columns=columns)
                    except Exception as e:
                        self.log.warning(f"Error reading previous VERSION sheet: {e}")
                    next_version = self.version
                else:
                    next_version = "1.0.0"
                    out_file = os.path.join(
                        self.output_dir,
                        f"Relecov_metadata_template_v{next_version}.xlsx",
                    )
            else:
                next_version = "1.0.0"
                out_file = os.path.join(
                    self.output_dir,
                    f"Relecov_metadata_template_v{next_version}.xlsx",
                )
            # Store versioning information
            version_info = {
                "FILE_VERSION": f"Relecov_metadata_template_v{next_version}",
                "CODE": next_version,
                "NOTES CONTROL": notes_control_input,
                "DATE": datetime.now().strftime("%Y-%m-%d"),
            }
            version_history = pd.concat(
                [version_history, pd.DataFrame([version_info])], ignore_index=True
            )
            out_file = os.path.join(
                self.output_dir, f"Relecov_metadata_template_v{next_version}.xlsx"
            )

            # Define required metadata classifications
            required_classification = [
                "Database Identifiers",
                "Sample collection and processing",
                "Host information",
                "Sequencing",
                "Pathogen diagnostic testing",
                "Contributor Acknowledgement",
                "Public databases",
                "Bioinformatics and QC metrics fields",
            ]
            required_properties = json_schema.get("required")
            schema_properties = json_schema.get("properties")

            # Read json schema properties and convert it into pandas df
            try:
                schema_properties_flatten = relecov_tools.assets.schema_utils.metadatalab_template.schema_to_flatten_json(
                    schema_properties
                )
                df = relecov_tools.assets.schema_utils.metadatalab_template.schema_properties_to_df(
                    schema_properties_flatten
                )
                # Filter metadata fields based on required classifications
                df = df[df["classification"].isin(required_classification)]
                df["required"] = df["property_id"].apply(
                    lambda x: "Y" if x in required_properties else "N"
                )

                # TODO: Define enums without ontologies for release 1.6.0
                # def clean_ontologies(enums):
                #     return [re.sub(r"\s*\[.*?\]", "", item).strip() for item in enums]

                # df["enum"] = df["enum"].apply(
                #     lambda enum_list: (
                #         clean_ontologies(enum_list)
                #         if isinstance(enum_list, list)
                #         else enum_list
                #     )
                # )
            except Exception as e:
                self.log.error(f"Error processing schema properties: {e}")
                stderr.print(f"Error processing schema properties: {e}")
                return None

            # Ensure 'header' column exists before filtering
            if "header" in df.columns:
                df["header"] = df["header"].astype(str).str.strip()
                df_filtered = df[df["header"].str.upper() == "Y"]
            else:
                self.log.warning(
                    "No se encontró la columna 'header', usando df sin filtrar."
                )
                df_filtered = df

            # Overview sheet
            try:
                overview_header = [
                    "Label name",
                    "Description",
                    "Group",
                    "Mandatory (Y/N)",
                    "Example",
                ]
                df_overview = pd.DataFrame(
                    columns=[col_name for col_name in overview_header]
                )
                df_overview["Label name"] = df_filtered["label"]
                df_overview["Description"] = df_filtered["description"]
                df_overview["Group"] = df_filtered["classification"]
                df_overview["Mandatory (Y/N)"] = df_filtered["required"]
                df_overview["Example"] = df_filtered["examples"].apply(
                    lambda x: x[0] if isinstance(x, list) else x
                )
            except Exception as e:
                self.log.error(f"Error creating overview sheet: {e}")
                stderr.print(f"Error creating overview sheet: {e}")
                return None

            # Create Metadata LAB sheet
            try:
                metadatalab_header = ["REQUERIDO", "EJEMPLOS", "DESCRIPCIÓN", "CAMPO"]
                df_metadata = pd.DataFrame(
                    columns=[col_name for col_name in metadatalab_header]
                )
                df_metadata["REQUERIDO"] = df_filtered["required"].apply(
                    lambda x: "YES" if str(x).upper() in ["Y", "YES"] else ""
                )
                df_metadata["EJEMPLOS"] = df_filtered["examples"].apply(
                    lambda x: x[0] if isinstance(x, list) else x
                )
                df_metadata["DESCRIPCIÓN"] = df_filtered["description"]
                df_metadata["CAMPO"] = df_filtered["label"]
                df_metadata = df_metadata.transpose()
            except Exception as e:
                self.log.error(f"Error creating MetadataLab sheet: {e}")
                stderr.print(f"[red]Error creating MetadataLab sheet: {e}")
                return None

            # Create Data Validation sheet
            try:
                datavalidation_header = ["EJEMPLOS", "DESCRIPCIÓN", "CAMPO"]
                df_hasenum = df[(pd.notnull(df.enum))]
                df_validation = pd.DataFrame(
                    columns=[col_name for col_name in datavalidation_header]
                )
                df_validation["tmp_property"] = df_hasenum["property_id"]
                df_validation["EJEMPLOS"] = df_hasenum["examples"].apply(
                    lambda x: x[0] if isinstance(x, list) else x
                )
                df_validation["DESCRIPCIÓN"] = df_hasenum["description"]
                df_validation["CAMPO"] = df_hasenum["label"]
            except Exception as e:
                self.log.error(f"Error creating DataValidation sheet: {e}")
                stderr.print(f"[red]Error creating DataValidation sheet: {e}")
                return None

            try:

                enum_dict = {property: [] for property in df_hasenum["property_id"]}
                enum_maxitems = 0
                # Populate the dictionary with flattened lists
                for key in enum_dict.keys():
                    enum_values = df_hasenum[df_hasenum["property_id"] == key][
                        "enum"
                    ].values
                    if enum_values.size > 0:
                        enum_list = enum_values[0]  # Extract the list
                        enum_dict[key] = enum_list  # Assign the list to the dictionary
                        if enum_maxitems < len(enum_list):
                            enum_maxitems = len(enum_list)
                    else:
                        enum_dict[key] = []

                # Reshape list dimensions based on enum length.
                for key in enum_dict.keys():
                    if len(enum_dict[key]) < enum_maxitems:
                        num_nas = enum_maxitems - len(enum_dict[key])
                        for _ in range(num_nas):
                            enum_dict[key].append("")

                new_df = pd.DataFrame(enum_dict)
                new_index = range(len(new_df.columns))
                new_df.reindex(columns=new_index)

                valid_index = df_validation["tmp_property"].values
                valid_transposed = df_validation.transpose()
                valid_transposed.columns = valid_index

                frames = [valid_transposed, new_df]
                df_validation = pd.concat(frames)
                df_validation = df_validation.drop(index=["tmp_property"])
            except Exception as e:
                self.log.error(f"Error processing enums and combining data: {e}")
                stderr.print(f"[red]Error processing enums and combining data: {e}")
                return None

            #  Replace NaN, Inf values with empty strings
            df_overview = (
                df_overview.replace([float("inf"), float("-inf")], "")
                .fillna("")
                .infer_objects()
            )
            df_metadata = (
                df_metadata.replace([float("inf"), float("-inf")], "")
                .fillna("")
                .infer_objects()
            )
            df_validation = (
                df_validation.replace([float("inf"), float("-inf")], "")
                .fillna("")
                .infer_objects()
            )

            # WRITE EXCEL
            try:
                writer = pd.ExcelWriter(out_file, engine="xlsxwriter")
                relecov_tools.assets.schema_utils.metadatalab_template.excel_formater(
                    df_overview,
                    writer,
                    "OVERVIEW",
                    out_file,
                    have_index=False,
                    have_header=False,
                )
                relecov_tools.assets.schema_utils.metadatalab_template.excel_formater(
                    df_metadata,
                    writer,
                    "METADATA_LAB",
                    out_file,
                    have_index=True,
                    have_header=False,
                )
                relecov_tools.assets.schema_utils.metadatalab_template.excel_formater(
                    df_validation,
                    writer,
                    "DATA_VALIDATION",
                    out_file,
                    have_index=True,
                    have_header=False,
                )
                version_history.to_excel(writer, sheet_name="VERSION", index=False)
                writer.close()
                self.log.info(
                    f"Metadata lab template successfuly created in: {out_file}"
                )
                stderr.print(
                    f"[green]Metadata lab template successfuly created in: {out_file}"
                )
            except Exception as e:
                self.log.error(f"Error writing to Excel: {e}")
                stderr.print(f"[red]Error writing to Excel: {e}")
                return None

            try:
                wb = openpyxl.load_workbook(out_file)
                ws_metadata = wb["METADATA_LAB"]
                ws_metadata.freeze_panes = self.configurables.get("freeze_panel", "D1")
                ws_metadata.delete_rows(5)
                relecov_tools.assets.schema_utils.metadatalab_template.create_condition(
                    ws_metadata, self.project_config, df_filtered
                )
                relecov_tools.assets.schema_utils.metadatalab_template.add_conditional_format_age_check(
                    ws_metadata, df_filtered
                )
                ws_dropdowns = (
                    wb.create_sheet("DROPDOWNS")
                    if "DROPDOWNS" not in wb.sheetnames
                    else wb["DROPDOWNS"]
                )

                for row in ws_dropdowns.iter_rows():
                    for cell in row:
                        cell.value = None

                for col_idx, (property_id, enum_values) in enumerate(
                    zip(df["property_id"], df["enum"]), start=1
                ):
                    if isinstance(enum_values, list) and len(enum_values) > 0:
                        start_row = 1
                        col_letter = openpyxl.utils.get_column_letter(col_idx)
                        for row_offset, value in enumerate(
                            enum_values, start=start_row
                        ):
                            ws_dropdowns[f"{col_letter}{row_offset}"] = value

                        dropdown_range_address = f"DROPDOWNS!${col_letter}${start_row}:${col_letter}${start_row + len(enum_values) - 1}"

                        col_letter_metadata = ws_metadata.cell(
                            row=4, column=col_idx + 1
                        ).column_letter
                        dropdown_range_metadata = (
                            f"{col_letter_metadata}5:{col_letter_metadata}1000"
                        )
                        dropdown = DataValidation(
                            type="list",
                            formula1=f"{dropdown_range_address}",
                            allow_blank=False,
                            showErrorMessage=True,
                        )
                        dropdown.error = "El valor ingresado no es válido. Seleccione un valor de la lista desplegable."
                        dropdown.errorTitle = "Valor no permitido"
                        dropdown.prompt = f"Select a value for {property_id}"
                        dropdown.promptTitle = "Value selection"

                        ws_metadata.add_data_validation(dropdown)
                        dropdown.add(dropdown_range_metadata)

                if "OVERVIEW" in wb.sheetnames:
                    ws_overview = wb["OVERVIEW"]
                    column_width = 35
                    for col in ws_overview.columns:
                        max_length = 0
                        column = col[0].column_letter  # Get the column name
                        for cell in col:
                            try:
                                if len(str(cell.value)) > max_length:
                                    max_length = len(cell.value)
                            except Exception:
                                pass
                        adjusted_width = max_length + 2
                        ws_overview.column_dimensions[column].width = min(
                            adjusted_width, column_width
                        )

                    # Enable text wrapping for the entire sheet
                    for row in ws_overview.iter_rows():
                        for cell in row:
                            cell.alignment = openpyxl.styles.Alignment(wrap_text=True)

                    # ws_overview.protection.sheet = True
                    # ws_overview.protection.password = self.configurables.get(
                    #     "protection.password", ""
                    # )

                    # if "DATA_VALIDATION" in wb.sheetnames:
                    # ws_data_validation = wb["DATA_VALIDATION"]
                    # ws_data_validation.protection.sheet = True
                    # ws_data_validation.protection.password = self.configurables.get(
                    #     "protection.password", ""
                    # )

                    # if "VERSION" in wb.sheetnames:
                    # ws_data_validation = wb["VERSION"]
                    # ws_data_validation.protection.sheet = True
                    # ws_data_validation.protection.password = self.configurables.get(
                    #     "protection.password", ""
                    # )

                    ws_version = wb["VERSION"]
                    column_widths = []

                    for col in ws_version.columns:
                        max_length = 0
                        for cell in col:
                            if len(str(cell.value)) > max_length:
                                max_length = len(cell.value)
                        adjusted_width = max_length + 2
                        column_widths.append(adjusted_width)

                    # Apply the calculated column width
                    for i, width in enumerate(column_widths):
                        ws_version.column_dimensions[
                            openpyxl.utils.get_column_letter(i + 1)
                        ].width = width

                ws_dropdowns.sheet_state = "hidden"
                # ws_dropdowns.protection.sheet = True
                # ws_dropdowns.protection.password = "password123"

                wb.save(out_file)
            except Exception as e:
                self.log.error(f"Error adding dropdowns: {e}")
                stderr.print(f"[red]Error adding dropdowns: {e}")
                return None
        except Exception as e:
            self.log.error(f"Error in create_metadatalab_excel: {e}")
            stderr.print(f"[red]Error in create_metadatalab_excel: {e}")
            return None

    def summarize_schema(self, json_schema):
        """
        Generate summary statistics for a JSON Schema and display it in tabular format.

        Args:
            json_schema (dict): The JSON Schema to analyze.

        Returns:
            None: Displays the table directly to stdout.
        """
        properties = json_schema.get("properties", {})
        # Initialize counters
        total_properties = len(properties)
        type_counts = {}
        enum_count = 0
        free_text_count = 0

        # Iterate over properties
        for _, prop_details in properties.items():
            prop_type = prop_details.get("type", "unknown")

            # Count types
            type_counts[prop_type] = type_counts.get(prop_type, 0) + 1

            # Count enum vs free text
            if "enum" in prop_details:
                enum_count += 1
            else:
                free_text_count += 1

        # Prepare summary data
        summary_data = {
            "Total Properties": [total_properties],
            "String Properties": [type_counts.get("string", 0)],
            "Integer Properties": [type_counts.get("integer", 0)],
            "Number Properties": [type_counts.get("number", 0)],
            "Boolean Properties": [type_counts.get("boolean", 0)],
            "Object Properties": [type_counts.get("object", 0)],
            "Array Properties": [type_counts.get("array", 0)],
            "Enum Properties": [enum_count],
            "Free Text Properties": [free_text_count],
        }
        summary_df = pd.DataFrame(summary_data)

        # Display summary using rich table (if available) or print raw
        try:
            relecov_tools.utils.display_dataframe_to_user(
                name="JSON Schema Summary", dataframe=summary_df
            )
        except AttributeError:
            print(summary_df.to_string(index=False))

        def shorten_path(path, max_length=50):
            """Shortens long paths by keeping first and last segments while adding ellipsis in the middle."""
            if len(path) <= max_length:
                return path  # No need to shorten

            parts = path.split(os.sep)  # Split into parts
            if len(parts) > 3:
                return os.sep.join(
                    [parts[0], "..."] + parts[-2:]
                )  # Keep first, last two, and replace middle
            return path  # If it's already short, return as is

        # Folder containing results
        outdir_data = {
            "Description": [
                "Output Folder",
                "New JSON Schema",
                "Old JSON Schema",
                "Schema Diff File",
                "Metadata Template File",
            ],
            "Path": [
                shorten_path(self.output_dir),
                shorten_path(f"{self.output_dir}/relecov_schema.json"),
                shorten_path(self.base_schema_path),
                shorten_path(f"{self.output_dir}/build_schema_diff.txt"),
                shorten_path(f"{self.output_dir}/Relecov_metadata_template_v*.xlsx"),
            ],
        }

        # Convert to DataFrame
        outdir_df = pd.DataFrame(outdir_data)

        # Display summary using rich table or print raw if unavailable
        try:
            relecov_tools.utils.display_dataframe_to_user(
                name="JSON Results Overview", dataframe=outdir_df
            )
        except AttributeError:
            print(outdir_df.to_string(index=False))

    def handle_build_schema(self):
        # Load xlsx database and convert into json format
        self.log.info("Start reading xlsx database")
        stderr.print("[white]Start reading xlsx database")
        database_dic = self.read_database_definition()

        # Verify current schema used by relecov-tools:
        base_schema_json = relecov_tools.utils.read_json_file(self.base_schema_path)
        if not base_schema_json:
            self.log.error("Couldn't find relecov base schema.)")
            stderr.print("[red]Couldn't find relecov base schema. Exiting...)")
            sys.exit(1)

        # Create schema draft template (leave empty to be prompted to list of available schema versions)
        schema_draft_template = self.create_schema_draft_template()

        # build new schema draft based on database definition.
        new_schema_json = self.build_new_schema(database_dic, schema_draft_template)

        # Verify new schema follows json schema specification rules.
        self.verify_schema(new_schema_json)

        # Compare base vs new schema and print/saves differences (--diff = True)
        self.get_schema_diff(base_schema_json, new_schema_json)

        # Saves new JSON schema
        self.save_new_schema(new_schema_json)

        # Create metadata lab template
        if self.non_interactive or relecov_tools.utils.prompt_yn_question(
            "Do you want to create a metadata lab file?:"
        ):
            self.create_metadatalab_excel(new_schema_json)

        # Return new schema
        return new_schema_json



#!/usr/bin/env python
import json
import rich.console
import os
import re
from datetime import datetime as dtime
import relecov_tools.utils
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class RelecovMetadata(BaseModule):
    def __init__(
        self,
        metadata_file=None,
        sample_list_file=None,
        output_dir=None,
        files_folder=None,
        **kwargs,
    ):
        super().__init__(output_dir=output_dir, called_module=__name__)
        self.log.info("Initiating read-lab-metadata process")
        self.sample_list_file = sample_list_file
        self.files_folder = files_folder

        if metadata_file is None:
            self.metadata_file = relecov_tools.utils.prompt_path(
                msg="Select the excel file which contains metadata"
            )
        else:
            self.metadata_file = metadata_file

        if not os.path.exists(self.metadata_file):
            self.log.error("Metadata file %s does not exist ", self.metadata_file)
            stderr.print(
                "[red] Metadata file " + self.metadata_file + " does not exist"
            )
            raise FileNotFoundError(f"Metadata file {self.metadata_file} not found")

        if sample_list_file is None:
            stderr.print("[yellow]No samples_data.json file provided")
            self.log.warning("No samples_data.json file provided")
            if not os.path.isdir(str(files_folder)):
                stderr.print("[red]No samples file nor valid files folder provided")
                self.log.error("No samples file nor valid files folder provided")
                raise FileNotFoundError(
                    "No samples file nor valid files folder provided"
                )
            self.files_folder = os.path.abspath(files_folder)

        if sample_list_file is not None and not os.path.exists(sample_list_file):
            self.log.error(
                "Sample information file %s does not exist ", sample_list_file
            )
            stderr.print("[red] Samples file " + sample_list_file + " does not exist")
            raise FileNotFoundError(
                "Sample information file %s does not exist ", sample_list_file
            )
        else:
            self.sample_list_file = sample_list_file

        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = output_dir

        config_json = ConfigJson(extra_config=True)

        # TODO: remove hardcoded schema selection
        relecov_schema = config_json.get_topic_data("json_schemas", "relecov_schema")
        relecov_sch_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "schema", relecov_schema
        )
        self.configuration = config_json
        self.institution_config = config_json.get_configuration("institutions_config")

        out_path = os.path.realpath(self.output_dir)
        self.lab_code = out_path.split("/")[-2]
        self.logsum = self.parent_log_summary(
            output_dir=self.output_dir, lab_code=self.lab_code, path=out_path
        )

        with open(relecov_sch_path, "r") as fh:
            self.relecov_sch_json = json.load(fh)

        try:
            relecov_tools.assets.schema_utils.jsonschema_draft.check_schema_draft(
                self.relecov_sch_json, "2020-12"
            )
        except Exception as e:
            self.log.error("JSON schema is not valid: %s", str(e))
            stderr.print(f"[red]Error: JSON schema is not valid.\n{str(e)}")
            raise

        self.label_prop_dict = {}

        for prop, values in self.relecov_sch_json["properties"].items():
            try:
                self.label_prop_dict[values["label"]] = prop
            except KeyError:
                self.log.warning("Property %s does not have 'label' attribute", prop)
                stderr.print(
                    "[orange]Property " + prop + " does not have 'label' attribute"
                )
                continue
        self.date = dtime.now().strftime("%Y%m%d%H%M%S")
        self.json_req_files = config_json.get_topic_data(
            "lab_metadata", "lab_metadata_req_json"
        )
        self.schema_name = self.relecov_sch_json["title"]
        self.schema_version = self.relecov_sch_json["version"]
        self.metadata_processing = config_json.get_topic_data(
            "sftp_handle", "metadata_processing"
        )
        self.samples_json_fields = config_json.get_topic_data(
            "lab_metadata", "samples_json_fields"
        )
        self.unique_sample_id = "sequencing_sample_id"

    def get_samples_files_data(self, clean_metadata_rows):
        """Include the fields that would be included in samples_data.json

        Args:
            clean_metadata_rows (list(dict)): Cleaned list of rows from metadata_lab.xlsx file

        Returns:
            j_data (dict(dict)): Dictionary where each key is the sample ID and the values are
            its file names, locations and md5
        """

        def safely_calculate_md5(file):
            """Check file md5, but return Not Provided if file does not exist"""
            self.log.info("Generating md5 hash for %s...", str(file))
            try:
                return relecov_tools.utils.calculate_md5(file)
            except IOError:
                return "Not Provided [SNOMED:434941000124101]"

        # The files are and md5file are supposed to be located together
        dir_path = self.files_folder
        md5_checksum_files = [
            os.path.join(dir_path, f) for f in os.listdir(dir_path) if "md5" in f
        ]
        if md5_checksum_files:
            skip_list = self.configuration.get_topic_data(
                "sftp_handle", "skip_when_found"
            )
            md5_dict = relecov_tools.utils.read_md5_checksum(
                file_name=md5_checksum_files[0], avoid_chars=skip_list
            )
        else:
            md5_dict = {}
            self.log.warning("No md5sum file found.")
            self.log.warning("Generating new md5 hashes. This might take a while...")
        j_data = {}
        no_fastq_error = "No R1 fastq file was given for sample %s in metadata"
        n = 0
        for sample in clean_metadata_rows:
            n += 1
            sample_id = str(sample.get(self.unique_sample_id))
            if not sample_id:
                if sample.get("collecting_lab_sample_id"):
                    sample_id = sample["collecting_lab_sample_id"]
                    self.unique_sample_id = "collecting_lab_sample_id"
                else:
                    sample_id = sample.get("sequence_file_R1", "").split(".")[0]
                    self.unique_sample_id = "sequence_file_R1"
            files_dict = {}
            r1_file = sample.get("sequence_file_R1")
            r2_file = sample.get("sequence_file_R2")
            if not r1_file:
                self.logsum.add_error(
                    sample=sample_id,
                    entry=no_fastq_error % sample_id,
                )
                j_data[sample_id] = files_dict
                continue
            r1_md5 = md5_dict.get(r1_file)
            r2_md5 = md5_dict.get(r2_file)
            files_dict["sequence_file_R1"] = r1_file
            files_dict["sequence_file_path_R1"] = dir_path
            batch_id = dir_path.split("/")[-1]
            logtxt = f"Setting batch_id to {batch_id} based on download dir: {dir_path}"
            stderr.print(f"[yellow]{logtxt}")
            self.log.info(logtxt)
            files_dict["batch_id"] = dir_path.split("/")[-1]
            if not os.path.exists(os.path.join(dir_path, r1_file)):
                self.logsum.add_error(
                    sample=sample_id, entry="Provided R1 file not found after download"
                )
                continue
            if r1_md5:
                files_dict["sequence_file_R1_md5"] = r1_md5
            else:
                files_dict["sequence_file_R1_md5"] = safely_calculate_md5(
                    os.path.join(dir_path, r1_file)
                )
            if r2_file:
                files_dict["sequence_file_R2"] = r2_file
                files_dict["sequence_file_path_R2"] = dir_path
                if not os.path.exists(os.path.join(dir_path, r2_file)):
                    self.logsum.add_error(
                        sample=sample_id,
                        entry="Provided R2 file not found after download",
                    )
                    continue
                if r2_md5:
                    files_dict["sequence_file_R2_md5"] = r2_md5
                else:
                    files_dict["sequence_file_R2_md5"] = safely_calculate_md5(
                        os.path.join(dir_path, r1_file)
                    )
            if sample_id in j_data:
                sample_id = "_".join([sample_id, str(n)])
            j_data[sample_id] = files_dict
        if not any(val for val in j_data.values()):
            errtxt = f"No files found for the samples in {dir_path}"
            self.logsum.add_error(entry=errtxt, sample=sample_id)
            stderr.print(f"[red]{errtxt}")
        try:
            samples_filename = "_".join(["samples_data", self.lab_code + ".json"])
            samples_filename = self.tag_filename(filename=samples_filename)
            file_path = os.path.join(self.output_dir, samples_filename)
            relecov_tools.utils.write_json_to_file(j_data, file_path)
        except Exception:
            self.log.error("Could not output samples_data.json file to output folder")
        return j_data

    def match_to_json(self, valid_metadata_rows):
        """Keep only the rows from samples present in the input file samples.json

        Args:
            valid_metadata_rows (list(dict)): List of rows from metadata_lab.xlsx file

        Returns:
            clean_metadata_rows(list(dict)): List of rows matching the samples in samples_data.json
            missing_samples(list(str)): List of samples not found in samples_data.json
        """
        missing_samples = []
        if not self.sample_list_file:
            logtxt = "samples_data.json not provided, all samples will be included"
            self.logsum.add_warning(entry=logtxt)
            return valid_metadata_rows, missing_samples
        else:
            samples_json = relecov_tools.utils.read_json_file(self.sample_list_file)
        clean_metadata_rows = []
        for row in valid_metadata_rows:
            sample_id = str(row[self.unique_sample_id]).strip()
            self.logsum.feed_key(sample=sample_id)
            if sample_id in samples_json.keys():
                clean_metadata_rows.append(row)
            else:
                log_text = "Sample in metadata but missing in downloaded samples file"
                self.logsum.add_warning(sample=sample_id, entry=log_text)
                missing_samples.append(sample_id)
        return clean_metadata_rows, missing_samples

    def adding_fixed_fields(self, m_data):
        """Include fixed data that are always the same for every sample"""
        p_data = self.configuration.get_topic_data("lab_metadata", "fixed_fields")
        organism_mapping = self.configuration.get_topic_data(
            "lab_metadata", "organism_mapping"
        )

        for idx in range(len(m_data)):
            organism = m_data[idx].get("organism", "")
            if organism in organism_mapping:
                m_data[idx]["tax_id"] = organism_mapping[organism]["tax_id"]
                m_data[idx]["host_disease"] = organism_mapping[organism]["host_disease"]
            else:
                m_data[idx]["tax_id"] = "Missing [LOINC:LA14698-7]"
                m_data[idx]["host_disease"] = "Missing [LOINC:LA14698-7]"
            for key, value in p_data.items():
                m_data[idx][key] = value
            m_data[idx]["schema_name"] = self.schema_name
            m_data[idx]["schema_version"] = self.schema_version
            m_data[idx]["submitting_institution_id"] = self.lab_code
        return m_data

    def adding_copy_from_other_field(self, m_data):
        """Add a new field with information based in another field."""
        p_data = self.configuration.get_topic_data(
            "lab_metadata", "required_copy_from_other_field"
        )
        for idx in range(len(m_data)):
            for key, value in p_data.items():
                m_data[idx][key] = m_data[idx][value]
        return m_data

    def adding_post_processing(self, m_data):
        """Add fields which values require post processing"""
        p_data = self.configuration.get_topic_data(
            "lab_metadata", "required_post_processing"
        )
        for idx in range(len(m_data)):
            for key, p_values in p_data.items():
                value = m_data[idx].get(key)
                if not value:
                    continue
                if value in p_values:
                    p_field, p_set = p_values[value].split("::")
                    m_data[idx][p_field] = p_set
                else:
                    # Check if key p_values should match only part of the value
                    for reg_key, reg_value in p_values.items():
                        if reg_key in value:
                            p_field, p_set = reg_value.split("::")
                            m_data[idx][p_field] = p_set

        return m_data

    def adding_ontology_to_enum(self, m_data):
        """Read the schema to get the properties enum and, for those fields
        which have an enum property value, replace the value for the one
        that is defined in the schema.
        """
        enum_dict = {}
        for prop, values in self.relecov_sch_json["properties"].items():
            enum_values = values.get("enum", [])
            ontologies_present = any(
                isinstance(enum, str) and re.search(r" \[\w+:.*\]$", enum)
                for enum in enum_values
            )
            if not ontologies_present:
                continue
            if "enum" in values:
                enum_dict[prop] = {}
                for enum in values["enum"]:
                    go_match = re.search(r"(.+) \[\w+:.*", enum)
                    if go_match:
                        enum_dict[prop][go_match.group(1)] = enum
                    else:
                        enum_dict[prop][enum] = enum
        ontology_errors = {}
        for idx in range(len(m_data)):
            for key, e_values in enum_dict.items():
                if key in m_data[idx]:
                    current_value = m_data[idx][key]
                    if re.search(r" \[\w+:.*\]$", current_value):
                        continue  # If already has ontology, do nothing.
                    if current_value in e_values:
                        m_data[idx][key] = e_values[current_value]
                    else:
                        sample_id = m_data[idx][self.unique_sample_id]
                        log_text = f"No ontology found for {current_value} in {key}"
                        self.logsum.add_warning(sample=sample_id, entry=log_text)
                        ontology_errors[key] = ontology_errors.get(key, 0) + 1
                        continue
        if len(ontology_errors) >= 1:
            stderr.print(
                "[red] No ontology could be added in:\n",
                "\n".join({f"{x} - {y} samples" for x, y in ontology_errors.items()}),
            )
            self.log.warning(
                "No ontology could be added in:\n%s",
                "\n".join(f"{x} - {y} samples" for x, y in ontology_errors.items()),
            )

        return m_data

    def process_from_json(self, m_data, json_fields):
        """Find the labels that are missing in the file to match the given schema."""
        map_field = json_fields["map_field"]
        col_name = self.relecov_sch_json["properties"].get(map_field).get("label")
        json_data = json_fields["j_data"]
        for idx in range(len(m_data)):
            sample_id = str(m_data[idx].get(self.unique_sample_id))
            if m_data[idx].get(map_field):
                # Remove potential ontology tags from value like [SNOMED:258500001]
                cleaned_key = re.sub(" [\[].*?[\]]", "", m_data[idx][map_field])
                try:
                    adding_data = {
                        k: v
                        for k, v in json_data[cleaned_key].items()
                        if k in json_fields["adding_fields"]
                    }
                    m_data[idx].update(adding_data)
                except KeyError as error:
                    clean_error = re.sub("[\[].*?[\]]", "", str(error.args[0]))
                    if str(clean_error).lower().strip() == "not provided":
                        log_text = (
                            f"Label {col_name} was not provided in sample "
                            + f"{sample_id}, auto-completing with Not Provided"
                        )
                        self.logsum.add_warning(sample=sample_id, entry=log_text)
                    else:
                        log_text = (
                            f"Unknown field value {error} for json data: "
                            + f"{str(col_name)} in sample {sample_id}. Skipped"
                        )
                        self.logsum.add_warning(sample=sample_id, entry=log_text)
                        continue
                    # TODO: Include Not Provided as a configuration field
                    fields_to_add = {
                        x: "Not Provided [SNOMED:434941000124101]"
                        for x in json_fields["adding_fields"]
                    }
                    m_data[idx].update(fields_to_add)
        return m_data

    def infer_file_format_from_schema(self, metadata):
        """Infer the file_format field based on the extension in sequence_file_R1,
        using enum values (with ontology) directly from the schema."""

        extension_map = {
            ".fastq": "FASTQ",
            ".fastq.gz": "FASTQ",
            ".fq": "FASTQ",
            ".fq.gz": "FASTQ",
            ".bam": "BAM",
            ".cram": "CRAM",
            ".fasta": "FASTA",
            ".fa": "FASTA",
        }

        file_format_enum = (
            self.relecov_sch_json["properties"].get("file_format", {}).get("enum", [])
        )
        keyword_to_enum = {}

        for item in file_format_enum:
            match = re.match(r"^([A-Z]+)", item)
            if match:
                keyword = match.group(1)
                keyword_to_enum[keyword] = item

        for row in metadata:
            r1_file = row.get("sequence_file_R1", "").lower()
            file_format_val = None
            for ext, keyword in extension_map.items():
                if r1_file.endswith(ext.lower()):
                    file_format_val = keyword_to_enum.get(keyword)
                    break

            if file_format_val:
                row["file_format"] = file_format_val
            else:
                row["file_format"] = "Not Provided [SNOMED:434941000124101]"

        return metadata

    def adding_fields(self, metadata):
        """Add information located inside various json file as fields"""

        for key, values in self.json_req_files.items():
            stderr.print(f"[blue]Processing {key}")
            self.log.info(f"Processing {key}")
            f_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "conf", values["file"]
            )
            values["j_data"] = relecov_tools.utils.read_json_file(f_path)
            metadata = self.process_from_json(metadata, values)
            stderr.print(f"[green]Processed {key}")
            self.log.info(f"Processed {key}")

        if self.institution_config:
            self.log.info("Updating laboratory code from institutions_config...")
        # Include Sample information data from sample json file
        stderr.print("[blue]Processing sample data file")
        self.log.info("Processing sample data file")
        s_json = {}
        # TODO: Change sequencing_sample_id for some unique ID used in RELECOV database
        s_json["map_field"] = self.unique_sample_id
        s_json["adding_fields"] = self.samples_json_fields
        if self.sample_list_file:
            s_json["j_data"] = relecov_tools.utils.read_json_file(self.sample_list_file)
        else:
            s_json["j_data"] = self.get_samples_files_data(metadata)
        if not s_json["j_data"]:
            self.log.warning(
                f"Samples file {self.sample_list_file} is empty. All samples will be included"
            )
            s_json["j_data"] = self.get_samples_files_data(metadata)
        batch_id = self.get_batch_id_from_data(list(s_json["j_data"].values()))
        # This will declare self.batch_id in BaseModule() which will be used later
        self.set_batch_id(batch_id)
        metadata = self.process_from_json(metadata, s_json)
        metadata = self.infer_file_format_from_schema(metadata)
        stderr.print("[green]Processed sample data file.")
        self.log.info("Processed sample data file.")
        return metadata

    def read_configuration_json_files(self):
        """Read json files defined in configuration lab_metadata_req_json
        property
        """
        c_files = {}
        for item, value in self.json_files.items():
            f_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "conf", value
            )
            c_files[item] = relecov_tools.utils.read_json_file(f_path)
        return c_files

    def read_metadata_file(self):
        """Reads the input metadata file from header row, changes the metadata heading
        with their property name values defined in schema. Convert the date columns
        value to the yyyy/mm/dd format. Return list of dicts with data
        """
        meta_sheet = self.metadata_processing.get("excel_sheet")
        header_flag = self.metadata_processing.get("header_flag")
        sample_id_col = self.metadata_processing.get("sample_id_col")
        self.alternative_heading = False
        try:
            ws_metadata_lab, heading_row_number = relecov_tools.utils.read_excel_file(
                self.metadata_file, meta_sheet, header_flag, leave_empty=False
            )
        except KeyError:
            self.alternative_heading = True
            alt_sheet = self.metadata_processing.get("alternative_sheet")
            header_flag = self.metadata_processing.get("alternative_flag")
            sample_id_col = self.metadata_processing.get("alternative_sample_id_col")
            logtxt = f"No excel sheet named {meta_sheet}. Using {alt_sheet}"
            stderr.print(f"[yellow]{logtxt}")
            self.log.error(f"{logtxt}")
            ws_metadata_lab, heading_row_number = relecov_tools.utils.read_excel_file(
                self.metadata_file, alt_sheet, header_flag, leave_empty=False
            )
        valid_metadata_rows = []
        included_sample_ids = []
        row_number = heading_row_number
        for row in ws_metadata_lab:
            row_number += 1
            property_row = {}
            try:
                sample_id = str(row[sample_id_col]).strip()
            except KeyError:
                self.logsum.add_error(entry=f"No {sample_id_col} found in excel file")
                continue
            # Validations on the sample_id
            if sample_id in included_sample_ids:
                log_text = f"Skipped duplicated sample {sample_id} in row {row_number}. Sequencing sample id must be unique"
                self.logsum.add_warning(entry=log_text)
                continue
            if not row[sample_id_col] or "Not Provided" in sample_id:
                if row.get("collecting_lab_sample_id"):
                    sample_id = row["collecting_lab_sample_id"]
                else:
                    sample_id = row.get("sequence_file_R1", "").split(".")[0]
                if not sample_id:
                    log_text = (
                        f"{sample_id_col} not provided in row {row_number}. Skipped"
                    )
                    self.logsum.add_error(entry=log_text)
                    stderr.print(f"[red]{log_text}")
                    continue
                else:
                    log_text = f"{sample_id_col} not provided for {sample_id}"
                    self.logsum.add_error(entry=log_text, sample=sample_id)
                    stderr.print(f"[red]{log_text}")
            included_sample_ids.append(sample_id)
            for key in row.keys():
                if header_flag in key:
                    continue
                value = row[key]
                # Omitting empty or not provided values
                if value is None or value == "" or "not provided" in str(value).lower():
                    log_text = f"{key} not provided for sample {sample_id}"
                    self.logsum.add_warning(sample=sample_id, entry=log_text)
                    continue
                # Get JSON schema type
                schema_key = self.label_prop_dict.get(key, key)
                schema_type = (
                    self.relecov_sch_json["properties"]
                    .get(schema_key, {})
                    .get("type", "string")
                )
                # Conversion of values according to expected type
                try:
                    value = relecov_tools.utils.cast_value_to_schema_type(
                        value, schema_type
                    )
                except (ValueError, TypeError) as e:
                    log_text = f"Type conversion error for {key} (expected {schema_type}): {value}. {str(e)}"
                    self.logsum.add_error(sample=sample_id, entry=log_text)
                    stderr.print(f"[red]{log_text}")
                    continue
                if "date" in key.lower():
                    # Check if date is a string. Format YYYY/MM/DD to YYYY-MM-DD
                    pattern = r"^\d{4}[-/.]\d{2}[-/.]\d{2}"
                    if isinstance(row[key], dtime):
                        row[key] = str(row[key].date())
                    elif re.match(pattern, str(row[key])):
                        row[key] = str(row[key]).replace("/", "-").replace(".", "-")
                        row[key] = re.match(pattern, row[key]).group(0)
                        value = row[key]
                    else:
                        try:
                            row[key] = str(int(float(str(row[key]))))
                            self.log.info(
                                "Date given as an integer. Understood as a year"
                            )
                        except (ValueError, TypeError):
                            log_text = f"Invalid date format in {key}: {row[key]}"
                            self.logsum.add_error(sample=sample_id, entry=log_text)
                            stderr.print(f"[red]{log_text} for sample {sample_id}")
                            continue
                elif "sample id" in key.lower():
                    if isinstance(row[key], float) or isinstance(row[key], int):
                        row[key] = str(int(row[key]))
                else:
                    if isinstance(row[key], float) or isinstance(row[key], int):
                        row[key] = str(row[key])
                if "date" not in key.lower() and isinstance(row[key], dtime):
                    logtxt = f"Non-date field {key} provided as date. Parsed as int"
                    self.logsum.add_warning(sample=sample_id, entry=logtxt)
                    row[key] = str(relecov_tools.utils.excel_date_to_num(row[key]))
                property_row[schema_key] = value
            valid_metadata_rows.append(property_row)
        return valid_metadata_rows

    def create_metadata_json(self):
        stderr.print("[blue]Reading Lab Metadata Excel File")
        valid_metadata_rows = self.read_metadata_file()
        stderr.print(f"[green]Processed {len(valid_metadata_rows)} valid metadata rows")
        self.log.info(f"Processed {len(valid_metadata_rows)} valid metadata rows")
        clean_metadata_rows, missing_samples = self.match_to_json(valid_metadata_rows)
        if missing_samples:
            num_miss = len(missing_samples)
            logtx = "%s samples not found in metadata: %s" % (num_miss, missing_samples)
            self.logsum.add_warning(entry=logtx)
            stderr.print(f"[yellow]{num_miss} samples missing:\n{missing_samples}")
        # Continue by adding extra information
        stderr.print("[blue]Including additional information")
        self.log.info("Including additional information")

        extended_metadata = self.adding_fields(clean_metadata_rows)
        stderr.print("[blue]Including post processing information")
        self.log.info("Including post processing information")
        extended_metadata = self.adding_post_processing(extended_metadata)
        extended_metadata = self.adding_copy_from_other_field(extended_metadata)
        extended_metadata = self.adding_fixed_fields(extended_metadata)
        completed_metadata = self.adding_ontology_to_enum(extended_metadata)
        if not completed_metadata:
            self.log.warning("Metadata was completely empty. No output file generated")
            stderr.print("Metadata was completely empty. No output file generated")
            return
        file_code = "_".join(["lab_metadata", self.lab_code]) + ".json"
        file_name = self.tag_filename(filename=file_code)
        stderr.print("[blue]Writting output json file")
        os.makedirs(self.output_dir, exist_ok=True)
        # Creating log summary
        self.parent_create_error_summary()
        file_path = os.path.join(self.output_dir, file_name)
        self.log.info("Writting output json file %s", file_path)
        relecov_tools.utils.write_json_to_file(completed_metadata, file_path)
        return True



import datetime
import json
import os
import re
import shutil
import sys
from collections import Counter, defaultdict

import rich.console
import relecov_tools.utils
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class PipelineManager(BaseModule):
    def __init__(
        self,
        input=None,
        templates_root=None,
        output_dir=None,
        config=None,
        folder_names=None,
    ):
        super().__init__(output_dir=output_dir, called_module=__name__)
        self.current_date = datetime.date.today().strftime("%Y%m%d")
        self.log.info("Initiating pipeline-manager process")
        if input is None:
            self.input_folder = relecov_tools.utils.prompt_path(
                msg="Select the folder which contains the fastq file of samples"
            )
        else:
            self.input_folder = input
        if not os.path.exists(self.input_folder):
            self.log.error("Input folder %s does not exist ", self.input_folder)
            stderr.print("[red] Input folder " + self.input_folder + " does not exist")
            sys.exit(1)
        if templates_root is None:
            self.templates_root = relecov_tools.utils.prompt_path(
                msg="Select the folder path which contains the templates"
            )
        else:
            self.templates_root = templates_root
        if not os.path.exists(self.templates_root):
            self.log.error("Template folder %s does not exist ", self.templates_root)
            stderr.print(
                "[red] Template folder " + self.templates_root + " does not exist"
            )
            sys.exit(1)
        if config is None:
            config = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "conf",
                "configuration.json",
            )
        if not os.path.exists(config):
            self.log.error("Pipeline config file %s does not exist ", config)
            stderr.print("[red] Pipeline config file " + config + " does not exist")
            sys.exit(1)
        conf_settings = relecov_tools.utils.read_json_file(config)
        try:
            config_data = conf_settings["pipeline_manager"]
        except KeyError:
            self.log.error("Invalid pipeline config file %s ", config)
            stderr.print("[red] Invalid pipeline config file " + config)
        required_conf = [
            "analysis_user",
            "analysis_group",
            "analysis_folder",
            "sample_stored_folder",
            "sample_link_folder",
            "doc_folder",
            "organism_config",
        ]
        missing_conf = [k for k in required_conf if k not in config_data]
        if missing_conf:
            self.log.error("Invalid pipeline config file. Missing %s", missing_conf)
            stderr.print(f"[red]Invalid pipeline config file. Missing {missing_conf}")
            sys.exit(1)
        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = output_dir

        # Create the output folder if not exists
        try:
            os.makedirs(self.output_dir, exist_ok=True)
        except OSError or FileExistsError as e:
            self.log.error("Unable to create output folder %s ", e)
            stderr.print("[red] Unable to create output folder ", e)
            sys.exit(1)
        self.folder_list = folder_names
        # Update the output folder with the current date and analysis name
        self.out_folder_namevar = f"{self.current_date}_{config_data['analysis_group']}_%s_{config_data['analysis_user']}"
        self.analysis_folder = config_data["analysis_folder"]
        self.copied_sample_folder = config_data["sample_stored_folder"]
        self.linked_sample_folder = config_data["sample_link_folder"]
        self.doc_folder = config_data["doc_folder"]
        self.organism_config = config_data["organism_config"]

    def get_latest_lab_folders(self, initial_date):
        """Get latest folder with the newest date
        Args:
            initial_date(datetime.date()): Starting date to search for
        Returns:
            lab_latest_folders: list of paths with the latest folders
            latest_date: latest date in the folders
        """
        lab_folders = [f.path for f in os.scandir(self.input_folder) if f.is_dir()]
        lab_latest_folders = {}
        latest_date = initial_date
        for lab_folder in lab_folders:
            existing_upload_folders = False
            last_folder_date = initial_date
            scan_folder = os.path.join(self.input_folder, lab_folder)
            lab_sub_folders = [f.path for f in os.scandir(scan_folder) if f.is_dir()]
            for lab_sub_folder in lab_sub_folders:
                f_name = os.path.basename(lab_sub_folder)
                sub_f_date = relecov_tools.utils.string_to_date(f_name)
                if not sub_f_date:
                    continue
                if sub_f_date.date() > last_folder_date:
                    last_folder_date = sub_f_date.date()
                    latest_folder_name = lab_sub_folder
                    existing_upload_folders = True
            if existing_upload_folders:
                lab_latest_folders[lab_folder] = {
                    "path": latest_folder_name,
                    "date": last_folder_date,
                }
                if last_folder_date > latest_date:
                    latest_date = last_folder_date
        # keep only folders with the latest date to process
        lab_latest_folders = [
            d["path"] for d in lab_latest_folders.values() if d["date"] == latest_date
        ]
        self.log.info("Latest date to process is %s", latest_date)
        stderr.print("[blue] Collecting samples from date ", latest_date)
        return lab_latest_folders, latest_date

    def join_valid_items(self, input_folder, folder_list=[], initial_date="20220101"):
        """Join validated metadata for the latest batches downloaded into a single one

        Args:
            input_folder (str): Folder to start the searching process.
            initial_date (str): Only search for folders newer than this date.
            folder_list (list(str)): Only retrieve folders with these basenames. Defaults to list()

        Returns:
            join_validate (list(dict)): List of dictionaries containing all the samples
            found in each validated_lab_metadata.json form the scanned folders
            latest_date (str): Latest batch date found in the scanned folder
        """
        if folder_list:
            folders_to_process = []
            # TODO: Change this for os.walk but with multithreading because its too slow
            lab_folders = [f.path for f in os.scandir(self.input_folder) if f.is_dir()]
            for lab_folder in lab_folders:
                full_path = os.path.join(input_folder, lab_folder)
                lab_subfolders = [
                    f.path for f in os.scandir(full_path) if f.path if f.is_dir()
                ]
                folders_to_process.extend(
                    [f for f in lab_subfolders if os.path.basename(f) in folder_list]
                )
            if not folders_to_process:
                raise FileNotFoundError("No folders found with the given names")
            last_folder = sorted(folder_list)[-1]
            try:
                latest_date = relecov_tools.utils.string_to_date(last_folder).date()
            except ValueError:
                self.log.error(
                    "Failed to get date from folder names. Using last mod date"
                )
                latest_date = max(
                    [relecov_tools.utils.get_file_date(f) for f in folders_to_process]
                ).date()
        else:
            folders_to_process, latest_date = self.get_latest_lab_folders(initial_date)
        join_validate = list()
        for folder in folders_to_process:
            lab_code = folder.split("/")[-2]
            self.log.info("Collecting samples for %s", str(lab_code))
            stderr.print(f"[blue] Collecting samples for {lab_code}")
            # fetch the validate file and get sample id and r1 and r2 file path
            validate_files = [
                os.path.join(folder, f)
                for f in os.listdir(folder)
                if f.startswith("validated_lab_metadata") and f.endswith(".json")
            ]
            if not validate_files:
                self.log.error(f"No validated json file found for {folder}. Skipped")
                continue
            elif len(validate_files) > 1:
                self.log.error("Found multiple validated files in %s. Skipped" % folder)
                continue
            with open(validate_files[0]) as fh:
                data = json.load(fh)
            join_validate.extend(data)
        self.log.info("Found a total of %s samples", str(len(join_validate)))
        stderr.print(f"[blue]Found a total of {len(join_validate)} samples")
        return join_validate, latest_date

    def copy_process(self, samples_data, output_dir):
        """Copies all the necessary samples files in the given samples_data list
        to the output folder. Also creates symbolic links into the link folder
        given in config_file.

        Args:
            samples_data (list(dict)): samples_data from self.create_samples_data()
            output_dir (str): Destination folder to copy files

        Returns:
            samp_errors (dict): Dictionary where keys are sequencing_sample_id and values
            the files that received an error while trying to copy.
        """
        samp_errors = {}
        links_folder = os.path.join(
            output_dir, self.analysis_folder, self.linked_sample_folder
        )
        os.makedirs(links_folder, exist_ok=True)
        for sample in samples_data:
            sample_id = sample["sequencing_sample_id"]
            # fetch the file extension
            ext_found = re.match(r".*(fastq.*|bam)", sample["sequence_file_path_R1"])
            if not ext_found:
                self.log.error("No valid file extension found for %s", sample_id)
                samp_errors[sample_id].append(sample["sequence_file_path_R1"])
                continue
            ext = ext_found.group(1)
            seq_r1_sample_id = sample["sequencing_sample_id"] + "_R1." + ext
            # copy r1 sequencing file into the output folder self.analysis_folder
            sample_raw_r1 = os.path.join(
                output_dir, self.copied_sample_folder, seq_r1_sample_id
            )
            self.log.info("Copying sample %s", sample)
            stderr.print("[blue] Copying sample: ", sample["sequencing_sample_id"])
            try:
                shutil.copy(sample["sequence_file_path_R1"], sample_raw_r1)
                # create simlink for the r1
                r1_link_path = os.path.join(links_folder, seq_r1_sample_id)
                r1_link_path_ori = os.path.join("../../RAW", seq_r1_sample_id)
                os.symlink(r1_link_path_ori, r1_link_path)
            except FileNotFoundError as e:
                self.log.error("File not found %s", e)
                samp_errors[sample_id] = []
                samp_errors[sample_id].append(sample["sequence_file_path_R1"])
                if "sequence_file_path_R2" in sample:
                    samp_errors[sample_id].append(sample["sequence_file_path_R2"])
                continue
            # check if there is a r2 file
            if "sequence_file_path_R2" in sample:
                seq_r2_sample_id = sample["sequencing_sample_id"] + "_R2." + ext
                sample_raw_r2 = os.path.join(
                    output_dir,
                    self.copied_sample_folder,
                    seq_r2_sample_id,
                )
                try:
                    shutil.copy(sample["sequence_file_path_R2"], sample_raw_r2)
                    r2_link_path = os.path.join(links_folder, seq_r2_sample_id)
                    r2_link_path_ori = os.path.join("../../RAW", seq_r2_sample_id)
                    os.symlink(r2_link_path_ori, r2_link_path)
                except FileNotFoundError as e:
                    self.log.error("File not found %s", e)
                    if not samp_errors.get(sample_id):
                        samp_errors[sample_id] = []
                    samp_errors[sample_id].append(sample["sequence_file_path_R2"])
                    continue
        return samp_errors

    def create_samples_data(self, json_data):
        """Creates a copy of the json_data but only with relevant keys to copy files.
        Here 'sequence_file_path_R1' is created joining the original 'sequence_file_path_R1'
        and 'sequence_file_R1' fields. The same goes for 'sequence_file_path_R2'

        Args:
            json_data (list(dict)): Samples metadata in a list of dictionaries

        Returns:
            sample_data: list(dict)
                [
                  {
                    "sequencing_sample_id":XXXX,
                    "sequence_file_path_R1": XXXX,
                    "sequence_file_path_R2":XXXX
                  }
                ]
        """
        samples_data = []
        for item in json_data:
            sample = {}
            sample["sequencing_sample_id"] = item["sequencing_sample_id"]
            sample["sequence_file_path_R1"] = os.path.join(
                item["sequence_file_path_R1"], item["sequence_file_R1"]
            )
            if "sequence_file_path_R2" in item:
                sample["sequence_file_path_R2"] = os.path.join(
                    item["sequence_file_path_R2"], item["sequence_file_R2"]
                )
            samples_data.append(sample)
        return samples_data

    def split_data_by_key(self, json_data, keylist):
        """Split a given json data into different lists based on a given list of keys.
        From a single list of samples (dicts), the output will now be a list of lists
        where each new list is a subset of the original samples with the same values
        for the given list of keys

        Args:
            json_data (list(dict)): List of dictionaries, one for each sample
            keylist (list(str)): List of keys within the given dictionaries to
            split data.

        Returns:
            list_of_jsons_by_key (list(list(dict))): List of JSONs. One for each group
        """
        if not keylist:
            return [json_data]

        json_split_by_key = {}
        new_key = keylist[0]
        next_keys = keylist[1:]

        json_uniq_vals = frozenset([x.get(new_key) for x in json_data])
        for val in json_uniq_vals:
            grouped_samples = [x for x in json_data if x.get(new_key) == val]
            json_split_by_key[val] = grouped_samples

        list_of_jsons_by_key = []
        for group in json_split_by_key.values():
            list_of_jsons_by_key.extend(self.split_data_by_key(group, next_keys))
        return list_of_jsons_by_key

    def split_samples_by_organism(self, join_validate):
        """Split the input JSON into groups for each organism, which will later
        be analysed differently depending on its configuration.

        Args:
            join_validate (list(dict)): Merged JSON from all the ones found

        Returns:
            jsons_by_organism_dict(dict): Dictionary with {organism1:splitted_json1}
            with the result from split_data_by_key() for each group of samples from
            a certain organism.
        """
        jsons_by_organism_dict = {}
        for organism, org_conf in self.organism_config.items():
            org_samples = [x for x in join_validate if x.get("organism") == organism]
            if not org_samples:
                self.log.info(f"No samples found for organism: {organism}")
                continue
            if "group_by_fields" in org_conf.keys():
                logtxt = "Data for %s will be grouped by the following fields: %s"
                self.log.info(logtxt % (organism, str(org_conf["group_by_fields"])))
                keys_to_split = org_conf["group_by_fields"]
            else:
                self.log.warning(
                    f"No group_by_fields found for {organism}. Data won't be grouped"
                )
                keys_to_split = []
            splitted_json = self.split_data_by_key(org_samples, keys_to_split)
            jsons_by_organism_dict[organism] = splitted_json
        return jsons_by_organism_dict

    def process_samples(self, splitted_json, org_conf, latest_date):
        """Create the output folder, the required template for the pipeline and
        all the files necessary for each group of samples depending on the organism
        config given. Copy the samples and log the ones that could not be copied.

        Args:
            splitted_json (list(list)): List of jsons, one for each group
            org_conf (dict): organism params from configuration.json
            latest_date (str): Latest date found. Used to tag the analysis folder

        Returns:
            global_samp_errors(dict): Dictionary holding the samples with errors for
            each group
        """
        global_samp_errors = {}
        keys_to_split = org_conf.get("group_by_fields", [])
        # iterate over the sample_data to copy the fastq files in the output folder
        for idx, list_of_samples in enumerate(splitted_json, start=1):
            group_tag = f"{latest_date}_{org_conf['service_tag']}{idx:02d}"
            self.log.info("Processing group %s", group_tag)
            fields = {k: list_of_samples[0].get(k, "") for k in keys_to_split}
            if keys_to_split:
                logtxt = f"[blue]Processing group {group_tag} with fields: {fields}..."
                stderr.print(logtxt)
            else:
                logtxt = f"[blue]Processing group {group_tag}"
                stderr.print(logtxt)
            group_outfolder = os.path.join(
                self.output_dir, self.out_folder_namevar % group_tag
            )
            if os.path.exists(group_outfolder):
                msg = f"Analysis folder {group_outfolder} already exists and it will be deleted. Do you want to continue? Y/N"
                confirmation = relecov_tools.utils.prompt_yn_question(msg)
                if confirmation is False:
                    continue
                shutil.rmtree(group_outfolder)
                self.log.info(f"Folder {group_outfolder} removed")
            samples_data = self.create_samples_data(list_of_samples)
            # Create a folder for the group of samples and copy the files there
            self.log.info("Creating folder for group %s", group_tag)
            stderr.print(f"[blue]Creating folder for group {group_tag}")

            pipeline_templates = org_conf.get("pipeline_templates")
            if pipeline_templates is not None:
                if not isinstance(pipeline_templates, list):
                    raise ValueError(
                        f"'pipeline_templates' must be a list, but got {type(pipeline_templates).__name__}"
                    )
                for template_name in pipeline_templates:
                    template_path = os.path.join(self.templates_root, template_name)
                    if not os.path.exists(template_path):
                        self.log.warning(
                            f"Template {template_path} does not exist. Skipping."
                        )
                        continue
                    for item in os.listdir(template_path):
                        s = os.path.join(template_path, item)
                        d = os.path.join(group_outfolder, item)
                        if os.path.isdir(s):
                            shutil.copytree(s, d, dirs_exist_ok=True)
                        else:
                            shutil.copy2(s, d)
            else:
                if "pipeline_template" not in org_conf:
                    raise ValueError(
                        "Organism config must include either 'pipeline_templates' (list) or 'pipeline_template' (string)."
                    )
                template_name = org_conf["pipeline_template"]
                if not isinstance(template_name, str):
                    raise ValueError(
                        f"'pipeline_template' must be a string, but got {type(template_name).__name__}"
                    )
                template_path = os.path.join(self.templates_root, template_name)
                if not os.path.exists(template_path):
                    raise FileNotFoundError(f"Template {template_path} does not exist.")
                shutil.copytree(template_path, group_outfolder)

            # Check for possible duplicates
            self.log.info("Samples to copy %s", len(samples_data))
            # Extract the sequencing_sample_id from the list of dictionaries
            sample_ids = [item["sequencing_sample_id"] for item in samples_data]
            # Use Counter to count the occurrences of each sequencing_sample_id
            id_counts = Counter(sample_ids)
            # Find the sequencing_sample_id values that are duplicated (count > 1)
            duplicates = [
                sample_id for sample_id, count in id_counts.items() if count > 1
            ]
            if duplicates:
                self.log.error(
                    "Duplicate samples in group %s: %s" % (group_tag, duplicates)
                )
                stderr.print(
                    f"[red] There are duplicated samples in group {group_tag}: {duplicates}. Please handle manually"
                )
                continue
            samp_errors = self.copy_process(samples_data, group_outfolder)
            if len(samp_errors) > 0:
                stderr.print(
                    f"[red]Unable to copy files from {len(samp_errors)} samples in group {group_tag}"
                )
                msg = f"Do you want to delete analysis folder {group_outfolder}? Y/N"
                confirmation = relecov_tools.utils.prompt_yn_question(msg)
                if confirmation:
                    shutil.rmtree(group_outfolder)
                    self.log.info(f"Folder {group_outfolder} removed")
                    continue
            global_samp_errors[group_tag] = samp_errors
            samples_copied = len(list_of_samples) - len(samp_errors)
            copied_samps_log = f"Group {group_tag}: {samples_copied} samples copied out of {len(list_of_samples)}"
            self.log.info(copied_samps_log)
            stderr.print(copied_samps_log)

            dest_folder = os.path.join(group_outfolder, self.copied_sample_folder)

            samples_by_path = defaultdict(list)
            for s in samples_data:
                institution_path = os.path.dirname(
                    s.get("sequence_file_path_R1", "Missing [LOINC:LA14698-7]")
                )
                samples_by_path[institution_path].append(s["sequencing_sample_id"])

            errors_by_path = defaultdict(set)
            for sample_id in samp_errors:
                matching_samples = [
                    s for s in samples_data if s["sequencing_sample_id"] == sample_id
                ]
                if matching_samples:
                    institution_path = os.path.dirname(
                        matching_samples[0].get(
                            "sequence_file_path_R1", "Missing [LOINC:LA14698-7]"
                        )
                    )
                else:
                    institution_path = "Missing [LOINC:LA14698-7]"
                errors_by_path[institution_path].add(sample_id)

            for institution_path, samples in samples_by_path.items():
                total = len(samples)
                copied = total - len(errors_by_path.get(institution_path, []))
                msg = f"{copied}/{total} samples from {institution_path} copied to {dest_folder}"
                self.log.info(msg)
                stderr.print(msg)

            final_valid_samples = [
                x
                for x in list_of_samples
                if x.get("sequencing_sample_id") not in samp_errors
            ]
            sample_ids = [i for i in sample_ids if i not in samp_errors]
            group_analysis_folder = os.path.join(group_outfolder, self.analysis_folder)
            group_doc_folder = os.path.join(group_outfolder, self.doc_folder)
            # print samples_id file
            stderr.print(
                f"[blue]Generating sample_id.txt file in {group_analysis_folder}..."
            )

            with open(os.path.join(group_analysis_folder, "samples_id.txt"), "w") as f:
                for sample_id in sample_ids:
                    f.write(f"{sample_id}\n")
            group_info = os.path.join(group_doc_folder, "group_fields.json")
            relecov_tools.utils.write_json_to_file(fields, group_info)
            self.log.info(f"Group fields info saved in {group_info}")

            output_filename = self.tag_filename(f"{group_tag}_validate_batch.json")
            json_filename = os.path.join(group_doc_folder, output_filename)
            relecov_tools.utils.write_json_to_file(final_valid_samples, json_filename)
            self.log.info("Successfully created pipeline folder. Ready to launch")
            stderr.print(f"[blue]Folder {group_outfolder} finished. Ready to launch")
        return global_samp_errors

    def pipeline_exc(self):
        """Prepare folder for analysis in HPC
        Copies template selected as input
        Copies RAW data with sequencing id as fastq file names
        Creates samples_id.txt

        Args:

        Returns:

        """
        # collect json with all validated samples
        init_date = datetime.datetime.strptime("20220101", "%Y%m%d").date()
        join_validate, latest_date = self.join_valid_items(
            input_folder=self.input_folder,
            initial_date=init_date,
            folder_list=self.folder_list,
        )
        latest_date = str(latest_date).replace("-", "")
        if len(join_validate) == 0:
            stderr.print("[yellow]No samples were found. Aborting")
            sys.exit(0)
        batch_id = self.get_batch_id_from_data(join_validate)
        # If more than one batch is included, current date will be set as batch_id
        self.set_batch_id(batch_id)
        jsons_by_organism_dict = self.split_samples_by_organism(join_validate)
        if not jsons_by_organism_dict:
            self.log.error("No samples found for any of the organisms in config")
            raise ValueError("No samples found for any of the organisms in config")
        global_samp_errors = {}
        for organism, splitted_json in jsons_by_organism_dict.items():
            stderr.print(f"[blue]Processing samples for organism: {organism}")
            samp_errors = self.process_samples(
                splitted_json, self.organism_config[organism], latest_date
            )
            global_samp_errors[organism] = samp_errors

        for organism, org_samp_errors in global_samp_errors.items():
            if all(not v for v in org_samp_errors.values()):
                logtxt = f"All samples were copied successfully for {organism}!!"
                self.log.info(logtxt)
                stderr.print(f"[green]{logtxt}")
                continue
            else:
                self.log.error(f"Found errors during sample copying for {organism}: ")
            for group, samples in org_samp_errors.items():
                if not samples:
                    continue
                self.log.error(
                    "Group %s received error for samples: %s" % (group, samples)
                )
        self.log.info("Finished execution")
        stderr.print("Finished execution")
        return


class ResultUpload:
    def __init__(self, input_folder=None, conf_file=None):
        if input_folder is None:
            self.input_folder = relecov_tools.utils.prompt_path(
                msg="Select the folder which contains the results"
            )
        else:
            self.input_folder = input_folder
        if not os.path.exists(self.input_folder):
            self.log.error("Input folder %s does not exist ", self.input_folder)
            stderr.print("[red] Input folder " + self.input_folder + " does not exist")
            sys.exit(1)

        conf_file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "conf",
            "configuration.json",
        )
        if not os.path.exists(conf_file):
            self.log.error("Configuration file %s does not exist ", self.conf_file)
            stderr.print(
                "[red] Pipeline config file "
                + self.pipeline_conf_file
                + " does not exist"
            )
            sys.exit(1)
        conf_settings = relecov_tools.utils.read_json_file(conf_file)
        try:
            data = conf_settings["pipelines"]["relecov"]
        except KeyError:
            self.log.error("Invalid pipeline config file %s ", self.pipeline_conf_file)
            stderr.print(
                "[red] Invalid pipeline config file " + self.pipeline_conf_file
            )
        stderr.print(f"[blue] Configuration file loaded  {data}")



import logging
import os
import shutil
from datetime import datetime

import relecov_tools.utils
from relecov_tools.log_summary import LogSum
from relecov_tools.config_json import ConfigJson


class BaseModule:
    """Base module Class to handle logs for all modules in the package"""

    # These variables should only be activated via CLI using --hex-code
    _global_hex_code = None
    _cli_log_file = None
    _cli_log_path_param = None
    # This indicates that there is already a process functioning,
    # therefore you should create new log handlers instead of redirecting
    _active_process = False
    _current_version = None
    _cli_command = None

    def __init__(self, output_dir: str = None, called_module: str = None):
        """Set logs output path based on the module being executed

        Args:
            output_dir (str, optional): Output folder to save called module logs. Defaults to None.
            called_module (str, optional): Name of the module being executed. Defaults to None.
        """
        if BaseModule._active_process:
            self.log = logging.getLogger(f"{__class__.__module__}.{called_module}")
        else:
            self.log = logging.getLogger()
        self.log.propagate = True
        self.log.info(f"RELECOV-tools version {BaseModule._current_version}")
        self.log.info(f"CLI command executed: {BaseModule._cli_command}")
        if called_module is None:
            called_module = self.log.name
        self.called_module = called_module.replace("relecov_tools.", "").replace(
            "-", "_"
        )
        self.batch_id = "temp_id"
        self.basemod_date = datetime.today().strftime("%Y%m%d%H%M%S")
        config = ConfigJson(extra_config=True)
        logs_config = config.get_configuration("logs_config")
        if self.called_module in logs_config.get("modules_outpath", {}):
            output_dir = logs_config["modules_outpath"][self.called_module]
        else:
            if output_dir is None:
                output_dir = logs_config.get("default_outpath", "/tmp/relecov_tools")
        output_dir = os.path.realpath(output_dir)
        self.basemod_outdir = output_dir
        if BaseModule._global_hex_code is None:
            if BaseModule._cli_log_file:
                hex_folder = os.path.dirname(BaseModule._cli_log_file)
            else:
                hex_folder = output_dir
            self.hex = relecov_tools.utils.get_safe_hex(hex_folder)
            BaseModule._global_hex_code = self.hex
        else:
            self.hex = BaseModule._global_hex_code
        new_logname = self.tag_filename(self.called_module + ".log")
        self.final_log_path = os.path.join(output_dir, new_logname)
        if BaseModule._active_process is True:
            log_level = self.log.getEffectiveLevel()
            handler = BaseModule.set_log_handler(self.final_log_path, level=log_level)
            self.log.addHandler(handler)
        else:
            # First time this class is initialized, set root handler and move cli-log-file
            log_level = self.log.root.getEffectiveLevel()
            if not self.log.handlers:
                handler = BaseModule.set_log_handler(
                    self.final_log_path, level=log_level
                )
                self.log.addHandler(handler)
            self.redirect_logs(
                self.final_log_path, old_log_path=BaseModule._cli_log_file
            )

        self.base_logsum = None
        # Set this after the first module starts in case it calls other modules
        BaseModule._active_process = True

    @staticmethod
    def set_log_handler(log_filepath, level=logging.DEBUG, only_stream=False):
        """Create a custom log handler for logging"""
        if only_stream is True:
            log_fh = logging.StreamHandler()
        else:
            # Ensure the parent directory of the log file exists
            # If log_filepath includes a directory, create it to avoid FileHandler errors.
            dirpath = os.path.dirname(log_filepath)
            if dirpath:
                os.makedirs(dirpath, exist_ok=True)
            log_fh = logging.FileHandler(log_filepath, encoding="utf-8")
        log_fh.setLevel(level)
        log_fh.setFormatter(
            logging.Formatter(
                "[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"
            )
        )
        return log_fh

    def redirect_logs(
        self,
        new_log_path: str,
        old_log_path: str = None,
    ):
        """
        Move output log file to destination while keeping track of past logs.
        Keep folder destination if self.cli_log_path is activated.
        """
        if old_log_path is None:
            log_file = self.get_log_file()
        else:
            log_file = old_log_path
        if new_log_path is None:
            if self.base_logsum is not None:
                outdir = self.base_logsum.output_dir
                new_log_path = os.path.join(
                    outdir, self.tag_filename(self.called_module + ".log")
                )
            else:
                # Cannot redirect log file
                self.log.error(
                    "Could not redirect log file, no new destination nor active logsum"
                )
                return
        if log_file == new_log_path:
            self.log.debug(
                "log_file and new_log_path are the same, no redirection needed"
            )
            return
        if BaseModule._cli_log_path_param:
            self.log.warning(
                "--log-path activated via CLI. Logs output folder wont change"
            )
            new_log_path = os.path.join(
                BaseModule._cli_log_path_param, os.path.basename(new_log_path)
            )
        try:
            # Only copy file content, not permissions nor metadata
            shutil.copyfile(log_file, new_log_path)
            os.remove(log_file)
            self.log.debug(
                f"Successful redirection of old log-file {log_file} to {new_log_path}"
            )
        except OSError as e:
            self.log.error(f"Could not redirect {log_file} to {new_log_path}: {e}")
            return
        for handler in self.log.handlers:
            if isinstance(handler, logging.FileHandler):
                if self.called_module in handler.baseFilename:
                    self.log.debug(f"Removing handler for {handler.baseFilename}...")
                    self.log.removeHandler(handler)
                    handler.close()

        new_handler = BaseModule.set_log_handler(
            new_log_path, level=self.log.getEffectiveLevel()
        )
        self.log.addHandler(new_handler)
        self.log.debug(f"Redirected logs from {log_file} to {new_log_path}")
        return

    def get_log_file(self):
        """Retrieve the output path for the active logger"""
        module_handlers = []
        file_handlers = [
            x for x in self.log.handlers if isinstance(x, logging.FileHandler)
        ]
        for handler in file_handlers:
            if (
                self.called_module in handler.baseFilename
                and self.basemod_outdir in handler.baseFilename
            ):
                module_handlers.append(handler)
        if len(module_handlers) == 1:
            return module_handlers[0].baseFilename
        else:
            err_txt = "Could not retrieve log-file: "
            if len(module_handlers) == 0:
                self.log.error(
                    err_txt
                    + f"No output log-file defined for module {self.called_module}"
                )
            else:
                self.log.error(
                    err_txt
                    + f"Too many handlers defined for module {self.called_module}: {module_handlers}"
                )
            return ""

    def parent_log_summary(self, *args, **kwargs):
        """Initiate relecov_tools.LogSum class with given parameters"""
        if "output_dir" not in kwargs:
            kwargs["output_dir"] = self.basemod_outdir
        self.base_logsum = LogSum(**kwargs)
        return self.base_logsum

    def parent_create_error_summary(self, *args, **kwargs):
        """Output log summary in the same folder as log file if filepath is not given"""
        sum_filepath = self.final_log_path.replace(".log", "_log_summary.json")
        if "filepath" not in kwargs:
            kwargs["filepath"] = sum_filepath
        self.base_logsum.create_error_summary(**kwargs)
        return

    def set_batch_id(self, batch_id):
        """Set batch_id variable and rename log file to include this tag"""
        if self.batch_id == "temp_id":
            self.log.debug(f"Setting batch_id as {batch_id}...")
            self.batch_id = str(batch_id)
        self.final_log_path = self.tag_filename(self.final_log_path).replace(
            "_temp_id", ""
        )
        self.redirect_logs(self.final_log_path)
        return

    def tag_filename(self, filename: str):
        """Include hexadecimal code and batch_id in filename to avoid duplication"""
        tag = self.batch_id + "_" + self.hex
        base_name, extension = os.path.splitext(filename)
        for name, mark in {"batch_id": self.batch_id, "hex code": self.hex}.items():
            if mark in base_name:
                base_name = base_name.replace("_" + mark, "")
                self.log.debug(f"{filename} already includes {name}: {mark}.")
        return str(base_name) + "_" + tag + str(extension)

    def get_batch_id_from_data(self, json_data: list):
        """Extract batch_id from the first sample in JSON data"""
        if not json_data:
            self.log.debug("Provided json data is empty.")
            self.log.debug(f"Using execution date as batch_id: {self.basemod_date}")
            return self.basemod_date
        unique_batchs = {samp["batch_id"] for samp in json_data if "batch_id" in samp}
        if len(unique_batchs) == 0:
            filepath = "sequence_file_path_R1"
            self.log.debug(f"Found no samples with batch_id. Trying with {filepath}...")
            # If created with download module, batch_id will be the name of the folder
            unique_batchs = {
                samp[filepath].split("/")[-1] for samp in json_data if filepath in samp
            }
        if len(unique_batchs) == 1:
            batch_id = list(unique_batchs)[0]
            return batch_id
        elif len(unique_batchs) == 0:
            self.log.debug("Could not extract batch_id from provided json data")
        else:
            self.log.debug("More than one batch_id found in data.")
        self.log.debug(f"Using execution date as batch_id: {self.basemod_date}")
        batch_id = self.basemod_date
        return batch_id



#!/usr/bin/env python
import logging
import rich.console
import pandas as pd


import relecov_tools.utils
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.utils import column_index_from_string
from openpyxl.formatting.rule import FormulaRule
from openpyxl.styles import PatternFill


log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


def schema_to_flatten_json(json_data):
    """This function flattens schema when nested items are found"""
    try:
        flatten_json = {}
        for property_id, features in json_data.items():
            try:
                if features.get("type") == "array":
                    complex_properties = json_data[property_id]["items"].get(
                        "properties"
                    )
                    for (
                        complex_property_id,
                        complex_feature,
                    ) in complex_properties.items():
                        flatten_json.update({complex_property_id: complex_feature})
                else:
                    flatten_json.update({property_id: features})
            except Exception as e:
                stderr.print(f"[red]Error processing property {property_id}: {e}")
        return flatten_json
    except Exception as e:
        stderr.print(f"[red]Error in schema_to_flatten_json: {e}")
        return None


def schema_properties_to_df(json_data):
    try:
        # Initialize an empty list to store the rows of the DataFrame
        rows = []

        # Iterate over each property in the JSON data
        for property_id, property_features in json_data.items():
            try:
                # Create a dictionary to hold the property features
                row = {"property_id": property_id}
                row.update(property_features)

                # Append the row to the list of rows
                rows.append(row)
            except Exception as e:
                stderr.print(f"[red]Error processing property {property_id}: {e}")

        # Create a DataFrame from the list of rows
        df = pd.DataFrame(rows)

        # Return the DataFrame
        return df
    except Exception as e:
        stderr.print(f"[red]Error in schema_properties_to_df: {e}")
        return None


def excel_formater(df, writer, sheet, out_file, have_index=True, have_header=True):
    try:

        # Write the DataFrame to the specified sheet
        df.to_excel(
            writer, sheet_name=sheet, startrow=1, index=have_index, header=have_header
        )

        # Get the xlsxwriter workbook and worksheet objects.
        workbook = writer.book
        worksheet = writer.sheets[sheet]

        # Set up general column width
        worksheet.set_column(0, len(df.columns), 30)

        # General header format
        header_formater = workbook.add_format(
            {
                "bold": True,
                "text_wrap": False,
                "valign": "top",
                "fg_color": "#B9DADE",  # Light blue
                "border": 1,
                "locked": True,
            }
        )

        # Custom header format for METADATA_LAB (red text starting from column 2)
        red_header_formater = workbook.add_format(
            {
                "bold": True,
                "text_wrap": False,
                "valign": "top",
                "fg_color": "#B9DADE",  # Light blue background
                "color": "#f60606",  # Red text color
                "border": 1,
                "locked": True,
            }
        )

        # First column format
        first_col_formater = workbook.add_format(
            {
                "bold": True,
                "text_wrap": False,
                "valign": "center",
                "fg_color": "#B9DADE",  # Light blue
                "border": 1,
                "locked": True,
            }
        )

        cell_formater = workbook.add_format(
            {
                "border": 1,  # Apply border to every cell
                "locked": True,
            }
        )

        if sheet == "OVERVIEW":
            # Write the column headers with the defined format.
            for col_num, value in enumerate(df.columns.values):
                try:
                    worksheet.write(0, col_num, value, header_formater)
                except Exception as e:
                    stderr.print(f"Error writing header at column {col_num + 1}: {e}")

            # Write the first column with the defined format.
            for row_num in range(1, len(df) + 1):
                try:
                    worksheet.write(
                        row_num, 0, df.iloc[row_num - 1, 0], first_col_formater
                    )
                except Exception as e:
                    stderr.print(f"Error writing first column at row {row_num}: {e}")

            for row_num in range(1, len(df) + 1):
                for col_num in range(1, len(df.columns)):
                    try:
                        worksheet.write(
                            row_num,
                            col_num,
                            df.iloc[row_num - 1, col_num],
                            cell_formater,
                        )
                    except Exception as e:
                        stderr.print(
                            f"Error writing cell at row {row_num}, column {col_num}: {e}"
                        )

        if sheet == "METADATA_LAB" or sheet == "DATA_VALIDATION":
            # Write the column headers with the defined format.
            for col_num in range(0, len(df.columns)):
                for row_num in range(0, len(df)):
                    if row_num < 4:
                        try:
                            worksheet.write(
                                row_num,
                                col_num + 1,
                                df.iloc[row_num, col_num],
                                header_formater,
                            )
                        except Exception as e:
                            stderr.print(
                                f"Error writing first column at row {row_num}: {e}"
                            )
                        if row_num == 0 and col_num >= 0 and sheet == "METADATA_LAB":
                            try:
                                worksheet.write(
                                    row_num,
                                    col_num + 1,
                                    df.iloc[row_num, col_num],
                                    red_header_formater,
                                )
                            except Exception as e:
                                stderr.print(
                                    f"Error writing first row at column {col_num}: {e}"
                                )
            # Write the first column with the defined format.
            for index_num, index_val in enumerate(df.index):
                try:
                    worksheet.write(index_num, 0, index_val, first_col_formater)
                except Exception as e:
                    stderr.print(f"Error writing first column at row {row_num}: {e}")
    except Exception as e:
        stderr.print(f"Error in excel_formater: {e}")


def create_condition(ws_metadata, conditions, df_filtered):
    """This function creates conditions on METADATA_LAB template sheet"""
    label_to_property = dict(zip(df_filtered["label"], df_filtered["property_id"]))
    column_map = {}

    for cell in ws_metadata[4]:
        property_id = label_to_property.get(cell.value)
        if property_id in conditions:
            column_map[property_id] = cell.column_letter

    for property_id, rules in conditions.items():
        col_letter = column_map.get(property_id)
        if col_letter:
            start_row = rules.get("header_row_idx", 5)
            end_row = rules.get("max_rows", 1000)
            validation_type = rules.get("validation_type")
            formula1 = rules.get("formula1", "").replace("{col_letter}", col_letter)
            formula2 = rules.get("formula2")
            error_message = rules.get("error_message", "Valor no permitido")
            error_title = rules.get("error_title", "Error")

            cell_range = f"{col_letter}{start_row}:{col_letter}{end_row}"

            if validation_type:
                validation = DataValidation(
                    type=validation_type,
                    operator=rules.get("operator", "between"),
                    formula1=formula1,
                    formula2=formula2,
                    showErrorMessage=True,
                )
                validation.error = error_message
                validation.errorTitle = error_title
                ws_metadata.add_data_validation(validation)
                validation.add(cell_range)

            if rules.get("format_cells_as_date", False):
                col_idx = column_index_from_string(col_letter)
                for row in ws_metadata.iter_rows(
                    min_row=start_row, max_row=end_row, min_col=col_idx, max_col=col_idx
                ):
                    for cell in row:
                        cell.number_format = "yyyy-mm-dd"
    return ws_metadata


def add_conditional_format_age_check(
    ws_metadata,
    df_filtered,
    prop1="host_age_years",
    prop2="host_age_months",
    start_row=5,
    end_row=1000,
):
    """
    Applies a conditional formatting that marks in red if both cells are filled in
    (host_age_years and host_age_months) in the same row.
    """

    label_to_property = dict(zip(df_filtered["label"], df_filtered["property_id"]))

    column_map = {}
    for cell in ws_metadata[4]:
        property_id = label_to_property.get(cell.value)
        if property_id:
            column_map[property_id] = cell.column_letter

    col1_letter = column_map.get(prop1)
    col2_letter = column_map.get(prop2)
    if not col1_letter or not col2_letter:
        return

    red_fill = PatternFill(start_color="F54627", end_color="F54627", fill_type="solid")
    for row in range(start_row, end_row + 1):
        formula = f"=AND(NOT(ISBLANK(${col1_letter}{row})), NOT(ISBLANK(${col2_letter}{row})))"
        rule = FormulaRule(formula=[formula], fill=red_fill)

        for col in [col1_letter, col2_letter]:
            ws_metadata.conditional_formatting.add(f"{col}{row}", rule)



import logging
import sys
import rich.console

import relecov_tools.utils
import jsonschema
from jsonschema import Draft202012Validator

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)

SCHEMA_VALIDATORS = {
    "2020-12": Draft202012Validator,
}


def check_valid_version(draft_version):
    """Validate the provided draft version against available JSON Schema versions."""
    available_schemas = [version for version in SCHEMA_VALIDATORS.keys()]
    try:
        if not draft_version:
            draft_version = relecov_tools.utils.prompt_selection(
                "Choose a Json Schema valid version:", available_schemas
            )
        elif draft_version not in available_schemas:
            draft_version = relecov_tools.utils.prompt_selection(
                f"Draft version '{draft_version}' not found. Choose a valid Json Schema version:",
                available_schemas,
            )
    except Exception as e:
        stderr.print(f"[red]An error occurred while selecting the draft version: {e}")
        sys.exit(1)

    if not draft_version:
        stderr.print("[red]No valid draft version selected. Exiting.")
        sys.exit(1)

    stderr.print(f"[green]Using draft version: {draft_version}")
    return draft_version


def create_draft(draft_version, required_items=None):
    """Creates a JSON Schema Draft template with required fields."""

    # Get parameters to set create the schema
    url_str = SCHEMA_VALIDATORS[draft_version].META_SCHEMA["$id"]

    # Construct the draft template
    draft_template = {
        "$schema": url_str,
        "$id": "",
        "title": "",
        "description": "",
        "version": "",
        "type": "object",
        "properties": {},
    }

    # Include required fields if specified
    if required_items:
        draft_template["required"] = []

    return draft_template


def check_schema_draft(schema_draft, draft_version):
    """Validates the schema_draft against the JSON Schema Draft 2020-12 meta-schema."""
    if draft_version not in SCHEMA_VALIDATORS:
        stderr.print(f"[red]Unsupported draft version: {draft_version}")
        sys.exit(1)

    validator_class = SCHEMA_VALIDATORS[draft_version]

    try:
        validator_class.check_schema(schema_draft)
        stderr.print("[green]New schema is valid based on JSON Specification rules.")
    except jsonschema.ValidationError:
        stderr.print(f"[red] Json schema does not fulfill ${draft_version} Validation")
        promp_answ = relecov_tools.utils.prompt_yn_question(
            "Errors found during schema validation, proceed?:"
        )
        if not promp_answ:
            sys.exit(1)
    except Exception as e:
        stderr.print(f"[yellow]Error occurred during schema validation: {e}")
        promp_answ = relecov_tools.utils.prompt_yn_question(
            "Errors found during schema validation, proceed?:"
        )
        if not promp_answ:
            sys.exit(1)



import logging
import os
import smtplib
import relecov_tools.utils
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders
from jinja2 import Environment, FileSystemLoader


log = logging.getLogger(__name__)


class EmailSender:
    def __init__(self, config=None, template_path=None):
        self.config = config
        self.template_path = template_path
        self.yaml_cred_path = self.config.get("yaml_cred_path")

        if not self.config:
            raise ValueError("Configuration not loaded correctly.")

        if not os.path.exists(self.template_path):
            raise FileNotFoundError(
                f"The template file could not be found in path {self.template_path}."
            )

    def get_institution_info(
        self, institution_code, institutions_file="institutions.json"
    ):
        """
        Load the institution's information from the JSON file.
        """
        institutions_file = self.config.get(
            "institutions_guide_path", "institutions_guide.json"
        )
        institutions_data = relecov_tools.utils.read_json_file(institutions_file)

        if institutions_data and institution_code in institutions_data:
            return institutions_data[institution_code]
        else:
            print(f"No information found for code {institution_code}")
            return None

    def render_email_template(
        self,
        additional_info="",
        invalid_count=None,
        submitting_institution_code=None,
        template_name=None,
        batch=None,
        password=None,
        zip_filename=None,
    ):

        institution_info = self.get_institution_info(submitting_institution_code)
        if not institution_info:
            raise ValueError("Error: Institution information could not be obtained.")

        institution_name = institution_info["institution_name"]

        template_vars_dict = {
            "submitting_institution": institution_name,
            "invalid_count": (
                invalid_count.get(submitting_institution_code, 0)
                if invalid_count
                else 0
            ),
            "additional_info": additional_info,
            "batch": batch,
            "password": password,
            "submitting_institution_code": submitting_institution_code,
            "zip_filename": zip_filename,
        }

        templates_base_dir = os.path.dirname(self.template_path)
        env = Environment(loader=FileSystemLoader(templates_base_dir))
        template = env.get_template(template_name)
        email_template = template.render(**template_vars_dict)

        return email_template

    def send_email(self, receiver_email, subject, body, attachments, email_psswd):

        if not isinstance(receiver_email, list):
            raise ValueError(
                f"receiver_emails should be a list, but it received: {type(receiver_email)}"
            )

        if not all(isinstance(email, str) for email in receiver_email):
            raise ValueError("All elements in receiver_emails must be strings.")

        credentials = relecov_tools.utils.read_yml_file(self.yaml_cred_path)
        if not credentials:
            print("No credentials found.")
            return

        sender_email = self.config["email_host_user"]

        email_password = (
            email_psswd if email_psswd else credentials.get("email_password")
        )

        if not email_password:
            print("The e-mail password could not be found.")
            return

        default_cc = "bioinformatica@isciii.es"
        msg = MIMEMultipart()
        msg["From"] = sender_email
        msg["To"] = ", ".join(receiver_email)
        msg["CC"] = default_cc
        msg["Subject"] = subject
        msg.attach(MIMEText(body, "plain"))

        for attachment in attachments:
            with open(attachment, "rb") as attachment_file:
                part = MIMEBase("application", "octet-stream")
                part.set_payload(attachment_file.read())
                encoders.encode_base64(part)
                part.add_header(
                    "Content-Disposition",
                    f"attachment; filename={os.path.basename(attachment)}",
                )
                msg.attach(part)
        all_recipients = receiver_email + [default_cc]
        try:
            server = smtplib.SMTP(self.config["email_host"], self.config["email_port"])
            server.starttls()
            server.login(sender_email, email_password)
            server.sendmail(sender_email, all_recipients, msg.as_string())
            server.quit()
            log.info("Mail sent successfully.")
            print("Mail sent successfully.")
        except smtplib.SMTPException as e:
            log.error(f"Error sending the mail to {receiver_email}: {e}")



#!/usr/bin/env python
"""
Common utility function used for relecov_tools package.
"""
import os
import re
import sys
import yaml
import logging
import rich
import json

from pathlib import Path
from datetime import datetime
from Bio import SeqIO

from relecov_tools.config_json import ConfigJson
from relecov_tools.read_bioinfo_metadata import BioinfoReportLog
import relecov_tools.utils

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


# INIT Class
class LongTableParse:
    """
    - parse_a_list_of_dictionaries() : returns generated_JSON
    - saving_file(generated_JSON)
    - parsing_csv() : It manages all this proccess:
        - calling first to parse_a_list_of_dictionaries() and then calling to saving_file()
    """

    def __init__(self, file_path=None, pipeline_name=None, output_folder=None):
        if file_path is None:
            self.file_path = relecov_tools.utils.prompt_path(
                msg="Select the csv file which contains variant long table information"
            )
        else:
            self.file_path = file_path

        if not os.path.exists(self.file_path):
            log.error("Variant long table file %s does not exist ", self.file_path)
            stderr.print(
                f"[red] Variant long table file {self.file_path} does not exist"
            )
            sys.exit(1)

        if not self.file_path.endswith(".csv"):
            log.error("Variant long table file %s is not a csv file ", self.file_path)
            stderr.print(
                f"[red] Variant long table file {self.file_path} must be a csv file"
            )
            sys.exit(1)

        if output_folder is None:
            use_default = relecov_tools.utils.prompt_yn_question("Use default path?: ")
            if use_default:
                self.output_folder = os.getcwd()
            else:
                self.output_folder = relecov_tools.utils.prompt_path(
                    msg="Select the output folder:"
                )
        else:
            self.output_folder = output_folder
        Path(self.output_folder).mkdir(parents=True, exist_ok=True)

        json_file = os.path.join(
            os.path.dirname(__file__), "..", "..", "conf", "bioinfo_config.json"
        )
        config_json = ConfigJson(json_file)
        self.software_config = config_json.get_configuration(pipeline_name)
        self.long_table_heading = self.software_config["variants_long_table"]["content"]

    def validate_file(self, heading):
        """Check if long table file has all mandatory fields defined in
        configuration file
        """
        for field in self.long_table_heading:
            if field not in heading:
                log.error("Incorrect format file. %s is missing", field)
                stderr.print(f"[red]Incorrect Format. {field} is missing in file")
                sys.exit(1)
        return True

    def parse_file(self):
        """This function generates a json file from the csv file entered by
        the user (long_table.csv).
        Validate the file by checking the header line
        """
        with open(self.file_path, encoding="utf-8-sig") as fh:
            lines = fh.readlines()

        heading_index = {}
        headings_from_csv = lines[0].strip().split(",")
        for heading in self.long_table_heading.values():
            heading_index[heading] = headings_from_csv.index(heading)

        stderr.print("[green]\tSuccessful checking heading fields")
        log.info("Successful checking heading fields")

        samp_dict = {}
        for line in lines[1:]:
            line_s = line.strip().split(",")

            sample = line_s[heading_index["SAMPLE"]]
            if sample not in samp_dict:
                samp_dict[sample] = []

            variant_dict = {
                key: (
                    {key2: line_s[heading_index[val2]] for key2, val2 in value.items()}
                    if isinstance(value, dict)
                    else line_s[heading_index[value]]
                )
                for key, value in self.long_table_heading.items()
            }

            if re.search("&", line_s[heading_index["GENE"]]):
                # Example
                # 215184,NC_045512.2,27886,AAACGAACATGAAATT,A,PASS,1789,1756,1552,0.87,ORF7b&ORF8,gene_fusion,n.27887_27901delAACGAACATGAAATT,.,.,ivar,B.1.1.318
                # This only occurs (for now) as gene fusion, so we just duplicate lines with same values
                genes = re.split("&", line_s[heading_index["GENE"]])
                for gene in genes:
                    variant_dict_copy = variant_dict.copy()
                    variant_dict_copy["Gene"] = gene
                    samp_dict[sample].append(variant_dict_copy)
            else:
                samp_dict[sample].append(variant_dict)
        stderr.print("[green]\tSuccessful parsing data")
        log.info("Successful parsing long table data")
        return samp_dict

    def convert_to_json(self, samp_dict):
        j_list = []
        # Grab date from filename
        result_regex = re.search(
            "variants_long_table(?:_\d{14})?\.csv", os.path.basename(self.file_path)
        )
        if result_regex is None:
            stderr.print(
                "[red]\tWARN: Couldn't find variants long table file. Expected file name is:"
            )
            stderr.print(
                "[red]\t\t- variants_long_table.csv or variants_long_table_YYYYMMDD.csv. Aborting..."
            )
            sys.exit(1)
        else:
            analysis_date = relecov_tools.utils.get_file_date(self.file_path)
            if len(analysis_date) == 8:
                analysis_date = datetime.strptime(analysis_date, "%Y%m%d").strftime(
                    "%Y-%m-%d"
                )
        for key, values in samp_dict.items():
            j_dict = {"sample_name": key, "bioinformatics_analysis_date": analysis_date}
            j_dict["variants"] = values
            j_list.append(j_dict)
        return j_list

    def save_to_file(self, j_list, file_tag):
        """Transform the parsed data into a json file"""
        file_name = "long_table_" + file_tag + ".json"
        file_path = os.path.join(self.output_folder, file_name)
        if os.path.exists(file_path):
            stderr.print(
                f"[blue]Long table {file_path} file already exists. Merging new data if possible."
            )
            log.info(
                "Long table %s file already exists. Merging new data if possible."
                % file_path
            )
            original_table = relecov_tools.utils.read_json_file(file_path)
            samples_indict = {item["sample_name"]: item for item in original_table}
            for item in j_list:
                sample_name = item["sample_name"]
                if sample_name in samples_indict:
                    if samples_indict[sample_name] != item:
                        stderr.print(
                            f"[red]Same sample {sample_name} has different data in both long tables."
                        )
                        log.error(
                            "Sample %s has different data in %s and new long table. Can't merge."
                            % (sample_name, file_path)
                        )
                        return None
                else:
                    original_table.append(item)
            try:
                with open(file_path, "w") as fh:
                    fh.write(json.dumps(original_table, indent=4))
                stderr.print(
                    "[green]\tParsed data successfully saved to file:", file_path
                )
                log.info("Parsed data successfully saved to file: %s", file_path)
            except Exception as e:
                stderr.print("[red]\tError saving parsed data to file:", str(e))
                log.error("Error saving parsed data to file: %s", e)
        else:
            try:
                with open(file_path, "w") as fh:
                    fh.write(json.dumps(j_list, indent=4))
                stderr.print(
                    "[green]\tParsed data successfully saved to file:", file_path
                )
                log.info("Parsed data successfully saved to file: %s", file_path)
            except Exception as e:
                stderr.print("[red]\tError saving parsed data to file:", str(e))
                log.error("Error saving parsed data to file: %s", e)

    def parsing_csv(self):
        """
        Function called when using the relecov-tools long-table-parse function.
        """
        # Parsing longtable file
        parsed_data = self.parse_file()
        j_list = self.convert_to_json(parsed_data)
        return j_list


# END of Class


def parse_long_table(files_list, file_tag, pipeline_name, output_folder=None):
    """File handler to retrieve data from long table files and convert it into a JSON structured format.
    This function utilizes the LongTableParse class to parse the long table data.
    Since this utility handles and maps data using a custom way, it returns None to be avoid being  transferred to method read_bioinfo_metadata.BioinfoMetadata.mapping_over_table().

    Args:
        files_list (list): A list of paths to long table files.

    Returns:
        None: Indicates that the function does not return any meaningful value.
    """
    method_name = f"{parse_long_table.__name__}"
    method_log_report = BioinfoReportLog()

    # Handling long table data
    if len(files_list) == 1:
        files_list_processed = files_list[0]
        if not os.path.isfile(files_list_processed):
            method_log_report.update_log_report(
                method_name, "error", f"{files_list_processed} given file is not a file"
            )
            sys.exit(method_log_report.print_log_report(method_name, ["error"]))

        long_table = LongTableParse(
            file_path=files_list_processed,
            pipeline_name=pipeline_name,
            output_folder=output_folder,
        )
        # Parsing long table data and saving it
        long_table_data = long_table.parsing_csv()
        # Saving long table data into a file
        long_table.save_to_file(long_table_data, file_tag)
        stderr.print("[green]\tProcess completed")
        log.info("Long table process completed")
    elif len(files_list) > 1:
        method_log_report.update_log_report(
            method_name,
            "warning",
            f"Found {len(files_list)} variants_long_table files. This version is unable to process more than one variants long table each time.",
        )
    # This needs to return none to avoid being parsed by method mapping-over-table
    return None


def extract_consensus_stats(files_list, file_tag, pipeline_name, output_folder=None):
    """File handler to parse consensus data (fasta) into JSON structured format.

    Args:
        files_list (list): A list with paths to condensus files.

    Returns:
        consensus_data_processed: A dictionary containing consensus data handled.
    """
    method_name = f"{extract_consensus_stats.__name__}"
    method_log_report = BioinfoReportLog()

    consensus_data_processed = {}
    missing_consens = []
    for consensus_file in files_list:
        sequence_names = []
        genome_length = 0
        try:
            record_fasta = SeqIO.parse(consensus_file, "fasta")
            for record in record_fasta:
                sequence_names.append(record.description)
                genome_length += len(record.seq)
        except FileNotFoundError as e:
            missing_consens.append(e.filename)
            continue
        sample_key = os.path.basename(consensus_file).split(".")[0]
        # Update consensus data for the sample key
        consensus_data_processed[sample_key] = {
            "sequence_name": ", ".join(sequence_names),
            "genome_length": genome_length,
            "sequence_filepath": os.path.dirname(consensus_file),
            "sequence_filename": sample_key,
            "sequence_md5": relecov_tools.utils.calculate_md5(consensus_file),
        }
    # Report missing consensus
    conserrs = len(missing_consens)
    if conserrs >= 1:
        method_log_report.update_log_report(
            method_name,
            "warning",
            f"{conserrs} samples missing in consensus file: {missing_consens}",
        )
        method_log_report.print_log_report(method_name, ["valid", "warning"])
    return consensus_data_processed


def get_software_versions_yml(files_list, file_tag, pipeline_name, output_folder=None):
    """File handler to parse software versions from yaml.

    Args:
        files_list (list): A list with paths to software version files.

    Returns:
        version_dict: Dictionary of software versions in the format:
              { software_id: {"software_version": X, "software_name": Y} }
    """

    method_name = f"{get_software_versions_yml.__name__}"
    method_log_report = BioinfoReportLog()

    version_dict = {}

    for file in files_list:
        if not os.path.isfile(file):
            print(f"[WARNING] Version file {file} not found. Skipping.")
            method_log_report.update_log_report(
                method_name, "warning", f"Version file {file} not found. Skipping."
            )
            continue

        with open(file, "r") as f:
            try:
                yml_data = yaml.safe_load(f)
            except yaml.YAMLError as e:
                print(f"[ERROR] Failed to parse YAML file {file}: {e}")
                method_log_report.update_log_report(
                    method_name, "warning", f"Failed to parse YAML file {file}: {e}"
                )
                continue

        for section in yml_data.values():
            for name, version in section.items():
                software_name = name.strip()
                software_version = (
                    version.strip() if isinstance(version, str) else str(version)
                )
                software_id = software_name.lower()
                if "/" not in software_id:
                    software_id = software_id.split()[0]
                version_dict[software_id] = {
                    "software_version": software_version,
                    "software_name": software_name,
                }

    return version_dict


def evaluate_qc_samples(
    data, thresholds, conditions, invert_operator, is_not_evaluable
):
    """
    Perform QC evaluation on a list of sample data dictionaries using provided threshold logic.

    Parameters:
    -----------
    data : list of dict
        Each dict represents a sample and contains metrics to be evaluated.
    thresholds : dict
        A dictionary where keys are metric names and values are tuples (operator, threshold).
    conditions : dict
        A dictionary where keys are metric names and values are condition functions.
    invert_operator : function
        Function that returns the inverse of a comparison operator.
    is_not_evaluable : function
        Function to detect if a metric value should be skipped.
    log_report : BioinfoReportLog
        Logger to record validation and warning messages.
    log : Logger
        Standard Python logger for logging events.
    method_name : str
        Name of the method for logging purposes.

    Returns:
    --------
    Tuple of:
    - data : List[Dict[str, Any]]
        The input list with added 'qc_test' and optional 'qc_failed' keys.
    - warning_messages : List[str]
        List of warning messages that occurred during evaluation.
    """
    warning_messages = []

    for sample in data:
        try:
            qc_status = "pass"
            failed_reasons = []

            for param, condition in conditions.items():
                value = sample.get(param)
                try:
                    if value is None or not condition(value):
                        if is_not_evaluable(value):
                            log.info(
                                "%s is not evaluable for %s in sample %s",
                                value,
                                param,
                                sample.get("sequencing_sample_id", "unknown"),
                            )
                            continue
                        qc_status = "fail"
                        op, th = thresholds[param]
                        inverted_op = invert_operator(op)
                        failed_reasons.append(f"({param} {inverted_op} {th})")
                except (TypeError, ValueError):
                    if is_not_evaluable(value):
                        continue
                    qc_status = "fail"
                    failed_reasons.append(f"({param} = {value} invalid)")
                    msg = (
                        f"Sample {sample.get('sequencing_sample_id', 'unknown')} "
                        f"has unevaluable value for {param}: {value}"
                    )
                    warning_messages.append(msg)

            sample["qc_test"] = qc_status
            if qc_status == "fail" and failed_reasons:
                sample["qc_failed"] = " -- ".join(failed_reasons)

            msg = (
                "valid",
                f"{sample.get('sequencing_sample_id', 'unknown')} evaluated: {qc_status}",
            )

        except (TypeError, ValueError, AttributeError) as e:
            sample["qc_test"] = "fail"
            sample["qc_failed"] = f"(evaluation_error = {str(e)})"
            sample_id = sample.get("sequencing_sample_id", "unknown")
            msg = f"Error evaluating sample {sample_id}: {e}"
            warning_messages.append(msg)

    return data, warning_messages



#!/usr/bin/env python
import logging
import json
import os
import inspect
import copy
import re

import openpyxl
from rich.console import Console
from datetime import datetime
from collections import OrderedDict
from relecov_tools.utils import rich_force_colors
import relecov_tools.utils
from relecov_tools.config_json import ConfigJson


log = logging.getLogger(__name__)
stderr = Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=rich_force_colors(),
)


class LogSum:
    def __init__(
        self,
        output_dir: str = None,
        lab_code: str = None,
        path: list = None,
    ):
        if output_dir is not None:
            if not os.path.isdir(str(output_dir)):
                try:
                    os.makedirs(output_dir, exist_ok=True)
                except IOError:
                    raise IOError(f"Logs output folder {output_dir} doesnt exist")
        else:
            log.info("No output_dir provided, selecting it from config...")
            config_json = ConfigJson(extra_config=True)
            logs_config = config_json.get_configuration("logs_config")
            output_dir = logs_config.get("default_outpath", "/tmp")

        log.info(f"Log summary outpath set to {output_dir}")
        self.output_dir = output_dir

        # Store new arguments for possible future use
        self.lab_code = lab_code
        self.path = path

        # Map legacy attributes to new arguments for compatibility
        self.lab_code = lab_code if lab_code else None
        self.path = path if path else None
        self.logs = {}
        return

    def feed_key(self, key=None, sample=None, path=None):
        """Run update_summary() with no entry nor log_type. Add a new empty key"""
        if self.lab_code:
            key = self.lab_code
        self.update_summary(
            log_type=None, key=key, entry=None, sample=sample, path=path
        )

    def add_error(self, entry, key=None, sample=None, path=None):
        """Run update_summary() with log_type as errors"""
        if self.lab_code:
            key = self.lab_code
        log.error(entry)
        self.update_summary(
            log_type="errors", key=key, entry=entry, sample=sample, path=path
        )
        return

    def add_warning(self, entry, key=None, sample=None, path=None):
        """Run update_summary() with log_type as warnings"""
        if self.lab_code:
            key = self.lab_code
        log.warning(entry)
        self.update_summary(
            log_type="warnings", key=key, entry=entry, sample=sample, path=path
        )
        return

    def update_summary(self, log_type, key, entry, sample=None, path=None):
        """Create a dictionary with a defined structure for each new key. Add the
        entry to the dictionary if it already exists. Add it to samples if its a sample

        Args:
            key (str): Name of the key holding the logs. A folder or a sample.
            log_type (str): Type of log being added. Either 'errors' or 'warnings'
            entry (str): Content message of the log.
            sample (str, optional): Name of a sample within key if the log is for it
            one sample instead of the whole key/folder. Defaults to None.
        """
        feed_dict = OrderedDict({"valid": True, "errors": [], "warnings": []})
        # Removing strange characters
        current_key = str(key).replace("./", "")
        entry, sample = (str(entry), str(sample))
        if current_key not in self.logs.keys():
            self.logs[current_key] = copy.deepcopy(feed_dict)
            self.logs[current_key]["samples"] = OrderedDict()
        if self.path:
            self.logs[current_key].update({"path": str(self.path)})
        if path is not None:
            self.logs[current_key].update({"path": str(path)})
        if log_type is None:
            if sample != "None" and sample not in self.logs[current_key]["samples"]:
                self.logs[current_key]["samples"][sample] = copy.deepcopy(feed_dict)
            return
        if sample == "None":
            self.logs[current_key][log_type].append(entry)
        else:
            if sample not in self.logs[current_key]["samples"].keys():
                self.logs[current_key]["samples"][sample] = copy.deepcopy(feed_dict)
            self.logs[current_key]["samples"][sample][log_type].append(entry)
        return

    def prepare_final_logs(self, logs):
        """Sets valid field to false if any errors were found for each key/sample

        Args:
            logs (dict): Custom dictionary of logs.

        Returns:
            logs: logs with updated valid field values
        """
        for key in logs.keys():
            if logs[key].get("errors"):
                logs[key]["valid"] = False
            if logs[key].get("samples") is not None:
                for sample in logs[key]["samples"].keys():
                    if logs[key]["samples"][sample]["errors"]:
                        logs[key]["samples"][sample]["valid"] = False
        return logs

    def merge_logs(self, logs_list, key_name=None):
        """Merge a multiple set of logs without losing information

        Args:
            logs_list (list(dict)): List of logs for different processes,
            logs should only include the actual records,
            key_name (str, Optional): Name of the final key holding the logs.
            If None, all the keys found in logs_list will be included.

        Returns:
            final_logs (dict): Merged list of logs into a single record
        """

        def add_new_logs(current_logs, new_logs):
            """Merge two logs including all its keys"""
            merged_logs = copy.deepcopy(current_logs)

            for key, values in new_logs.items():
                if key not in merged_logs:
                    merged_logs[key] = copy.deepcopy(values)
                else:
                    for field in ["errors", "warnings"]:
                        merged_logs[key].setdefault(field, [])
                        merged_logs[key][field].extend(values.get(field, []))
                        # Remove repeated elements
                        merged_logs[key][field] = list(
                            dict.fromkeys(merged_logs[key][field])
                        )
                    merged_logs[key].setdefault("samples", {})
                    for sample, vals in new_logs[key].get("samples", {}).items():
                        if sample not in merged_logs[key]["samples"].keys():
                            merged_logs[key]["samples"][sample] = vals
                        else:
                            merged_logs[key]["samples"][sample]["errors"].extend(
                                new_logs[key]["samples"][sample]["errors"]
                            )
                            merged_logs[key]["samples"][sample]["warnings"].extend(
                                new_logs[key]["samples"][sample]["warnings"]
                            )
                    merged_logs[key].setdefault("path", self.output_dir)
            return merged_logs

        if not logs_list:
            return {}
        if len(logs_list) == 1:
            if key_name is None:
                return logs_list[0]
            else:
                if key_name in logs_list[0]:
                    return logs_list[0]
                else:
                    return {}
        proc_logs_list = []
        for logs in logs_list:
            if key_name is None:
                proc_logs_list.append(logs)
            else:
                proc_logs_list.append({key_name: logs.get(key_name, {})})
        merged_logs = proc_logs_list[0]
        for idx, logs in enumerate(proc_logs_list[1:]):
            if not logs:
                err = f"Logs {idx} were empty. Check if key {key_name} is present"
                log.warning(err)
                continue
            try:
                merged_logs = add_new_logs(merged_logs, logs)
            except (TypeError, KeyError) as e:
                err = f"Could not add logs {idx} in list: {e}"
                stderr.print(f"[red]{err}")
                log.error(err)
        return merged_logs

    def create_logs_excel(self, logs, excel_outpath):
        """Create an excel file with logs information

        Args:
            logs (dict, optional): Custom dictionary of logs. Useful to create outputs
            excel_outpath (str): Path to output excel file
        """

        def reg_remover(string, pattern):
            """Remove annotation between brackets in logs message"""
            string = str(string)
            string = string.replace("['", "'").replace("']", "'").replace('"', "")
            string = re.sub(pattern, "", string)
            return string.strip()

        def feed_logs_to_excel(logs, excel_outpath):
            """Feed the data from logs into an excel file, creating different
            sheets depending on the provided list from configuration file"""

            workbook = openpyxl.Workbook()
            # TODO: Include these fields in configuration.json
            sheet_names_and_headers = {
                "Global Report": ["Lab_id", "Valid", "Errors", "Warnings"],
                "Samples Report": [
                    "Lab_id",
                    "Sample ID given for sequencing",
                    "Valid",
                    "Errors",
                ],
                "Other warnings": [
                    "Lab_id",
                    "Sample ID given for sequencing",
                    "Valid",
                    "Warnings",
                ],
            }
            for name, header in sheet_names_and_headers.items():
                new_sheet = workbook.create_sheet(name)
                new_sheet.append(header)
            regex = r"[\[\]]"  # Regex to remove lists brackets

            for key, logs in logs.items():
                if not logs.get("samples"):
                    log.warning(f"No samples found for key {key}")
                    try:
                        samples_logs = logs[key]["samples"]
                    except (KeyError, AttributeError) as e:
                        err = f"Could not convert log summary to excel for {key}: {e}"
                        stderr.print(f"[red]{err}")
                        log.error(err)
                        workbook.close()
                        return
                else:
                    samples_logs = logs.get("samples")
                if not samples_logs:
                    logs["Warnings"].append("No samples found to report")

                valid = logs.get("valid", False)
                warnings = logs.get("warnings", [])

                warnings_list = warnings if isinstance(warnings, list) else [warnings]
                truncated_warnings = []
                max_lenght = 250

                for warning in warnings_list:
                    warnings_str = str(warning)
                    if len(warnings_str) > max_lenght:
                        warnings_str = warnings_str[:max_lenght] + "..."
                    truncated_warnings.append(warnings_str)
                warnings_cleaned = "; ".join(truncated_warnings)

                errors_list = logs.get("errors", [])
                errors_list = (
                    errors_list if isinstance(errors_list, list) else [errors_list]
                )

                truncated_errors = []

                for err in errors_list:
                    err_str = reg_remover(str(err), regex)
                    if len(err_str) > max_lenght:
                        err_str = err_str[:max_lenght] + "..."
                    truncated_errors.append(err_str)

                errors_cleaned = "; ".join(truncated_errors)

                workbook["Global Report"].append(
                    [str(key), str(valid), errors_cleaned, warnings_cleaned]
                )

                regex = (
                    r"\[.*?\]"  # Regex to remove ontology annotations between brackets
                )

                for sample, slog in samples_logs.items():
                    clean_errors = []
                    for x in slog["errors"]:
                        err_str = reg_remover(str(x), regex)
                        if len(err_str) > max_lenght:
                            err_str = err_str[:max_lenght] + "..."
                        clean_errors.append(err_str)
                    error_row = [
                        str(key),
                        sample,
                        str(slog["valid"]),
                        "\n".join(clean_errors),
                    ]
                    workbook["Samples Report"].append(error_row)

                    clean_warngs = []
                    for x in slog["warnings"]:
                        war_str = reg_remover(str(x), regex)
                        if len(war_str) > max_lenght:
                            war_str = war_str[:max_lenght] + "..."
                        clean_warngs.append(war_str)
                    warning_row = [
                        str(key),
                        sample,
                        str(slog["valid"]),
                        "\n ".join(clean_warngs),
                    ]
                    workbook["Other warnings"].append(warning_row)

            # Adjusting the size of the columns in the excel file
            for name in sheet_names_and_headers.keys():
                relecov_tools.utils.adjust_sheet_size(workbook[name])
            del workbook["Sheet"]
            workbook.save(excel_outpath)
            stderr.print(f"[green]Successfully created logs excel in {excel_outpath}")
            return

        def translate_fields(samples_logs):
            # TODO Translate logs to spanish using a local translator model like deepl
            return

        if not os.path.exists(os.path.dirname(excel_outpath)):
            os.makedirs(os.path.dirname(excel_outpath), exist_ok=True)
            log.warning(
                "Given report outpath does not exist, created it automatically: %s",
                os.path.dirname(excel_outpath),
            )
        file_ext = os.path.splitext(excel_outpath)[-1]
        excel_outpath = excel_outpath.replace(file_ext, ".xlsx")

        feed_logs_to_excel(logs, excel_outpath)

        return

    def create_error_summary(
        self, called_module=None, filepath=None, logs=None, to_excel=False
    ):
        """Dump the log summary dictionary into a file with json format. If any of
        the 'errors' key is not empty, the parent key value 'valid' is set to false.

        Args:
            called_module (str, optional): Name of the module running this code.
            filename (str, optional): Name of the output file. Defaults to None.
            logs (dict, optional): Custom dictionary of logs. Useful to create outputs
            with selective information within all logs. Key names must remain the same.
            to_excel (bool, optional): Wether to output logs in excel format or not
        """
        if logs is None:
            logs = self.logs
        else:
            if not isinstance(logs, dict):
                log.error("Logs input must be a dict. No output file generated.")
                stderr.print("[red]Logs input must be a dict. No output file.")
                return
        if not called_module:
            traceback_functions = [
                f.function for f in inspect.stack() if "__main__.py" in f.filename
            ]
            if traceback_functions:
                called_module = traceback_functions[0]
            else:
                called_module = ""
        if not filepath:
            date = datetime.today().strftime("%Y%m%d%-H%M%S")
            filename = "_".join([date, called_module, "log_summary.json"])
            os.makedirs(self.output_dir, exist_ok=True)
            filepath = os.path.join(self.output_dir, filename)
        else:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
        if os.path.exists(filepath):
            log.info(f"{filepath} already exists, merging its content...")
            with open(filepath, "r", encoding="utf-8") as f:
                present_logs = json.load(f)
            if self.lab_code:
                merge_key = self.lab_code
            else:
                merge_key = None
            try:
                merged_logs = self.merge_logs(
                    key_name=merge_key, logs_list=[present_logs, logs]
                )
                logs = merged_logs
            except Exception as e:
                filepath = filepath.replace(".json", "_2.json")
                err = f"Could not merge logs: {e}. Saving logs in {filepath} instead"
                log.error(err)
        final_logs = self.prepare_final_logs(logs)
        with open(filepath, "w", encoding="utf-8") as f:
            try:
                f.write(
                    json.dumps(
                        final_logs, indent=4, sort_keys=False, ensure_ascii=False
                    )
                )
                stderr.print(f"Process log summary saved in {filepath}")
                if to_excel is True:
                    self.create_logs_excel(
                        final_logs, filepath.replace("log_summary", "report")
                    )
            except Exception as e:
                stderr.print(f"[red]Error exporting logs to file: {e}")
                log.error("Error exporting logs to file: %s", str(e))
                f.write(str(final_logs))
        return

    def rename_log_key(self, old_key, new_key):
        """Rename a key in the logs

        Args:
            old_key (str): Current key name
            new_key (str): New key name
        """
        if old_key in self.logs.keys():
            if new_key not in self.logs.keys():
                self.logs[new_key] = self.logs.pop(old_key)
            else:
                log.warning(
                    f"Could not rename logsum key {old_key}: {new_key} already in logs"
                )
        else:
            log.warning(f"Could not rename logsum key {old_key}: key not in logs")
        return

    @staticmethod
    def get_invalid_count(validation_logs):
        """
        Counts the number of invalid samples in the logs data by checking the `valid` field.

        Args:
            validation_logs (dict): Dictionary containing the validation logs.

        Returns:
            dict: Dictionary with entry_key as keys and counts of invalid samples as values.
        """
        invalid_counts = {}
        for entry_key, entry_value in validation_logs.items():
            if "samples" in entry_value:
                samples = entry_value["samples"]
                for sample_key, sample_value in samples.items():
                    if "valid" in sample_value and not sample_value["valid"]:
                        if invalid_counts.get(entry_key):
                            invalid_counts[entry_key] += 1
                        else:
                            invalid_counts[entry_key] = 1
        return invalid_counts
